from .FormElement import FormElement, calc_offset


class Button(FormElement):
    @classmethod
    def get_name_node_offset(cls, raw_data):
        return calc_offset([(3, 1), (2, 0)], raw_data)

    pass



from .ExternalDataProcessor import ExternalDataProcessor


class ExternalDataProcessor801(ExternalDataProcessor):
    version = '801'
    pass

    def encode_version(self):
        return [[
            [
                "106",
                "0"
            ]
        ]]

    def encode_root(self):
        return [[
            "2",
            self.header["file_uuid"]
        ]]



from .Template803 import Template803


class CommonTemplate803(Template803):
    @classmethod
    def get_decode_header(cls, header_data):
        return header_data[0][1][1]

    @classmethod
    def get_template_type(cls, header_data):
        return header_data[0][1][2]



from .FormElement import FormElement, check_count_element, calc_offset
from ....helper import FuckingBrackets, str_decode
from ....ext_exception import ExtException


class Group(FormElement):
    pass

    @classmethod
    def get_name_node_offset(cls, raw_data):
        return calc_offset([(3, 1), (1, 1), (2, 0)], raw_data)

    @classmethod
    def decode(cls, form, path, raw_data):
        try:
            size = check_count_element([
                (3, 1), (1, 1), (17, 2)
            ], raw_data)
        except Exception as err:
            raise ExtException(parent=err)
        if raw_data[0] == '22' and size < 20:
            raise FuckingBrackets(detail=cls.__name__)

        data = super().decode(form, path, raw_data)
        index = calc_offset([(3, 1), (1, 1), (17, 0)], raw_data)
        data['child'] = cls.decode_list(form, raw_data, index, f"{path}/{data['name']}")
        return data

    @classmethod
    def encode(cls, form, path, data):
        try:
            child = data['child']
        except KeyError:
            child = []
        raw_data = super().encode(form, path, data)
        if not child:
            return raw_data
        index = calc_offset([(3, 1), (1, 1), (17, 0)], raw_data)
        cls.encode_list(form, child, raw_data, index, f"{path}/{data['name']}")
        return raw_data



import os
import re
from base64 import b64decode, b64encode
from uuid import uuid4

from .. import helper
from ..ext_exception import ExtException
from ..metadata_types import MetaDataTypes


class MetaObject:
    version = '803'
    ext_code = {'obj': 0}
    encrypted_types = ['text', 'image']
    _obj_info = None
    _obj_name = None

    re_meta_data_obj = re.compile('^[^.]+\.json$')
    directive_1c_uncomment = re.compile('(?P<n>\\n)(?P<d>[#|&])')
    directive_1c_comment = re.compile('(?P<n>\\n)(?P<c>// v8unpack )(?P<d>[#|&])')

    def __init__(self, *, obj_name=None, options=None):
        self.title = obj_name if obj_name else \
            self._obj_name if self._obj_name else self.get_class_name_without_version()
        self.header = {}
        self.code = {}
        self.file_list = []
        self.options = options

    def get_options(self, name, default=None):
        try:
            return self.options[name]
        except (TypeError, KeyError):
            return default

    def get_decode_header(self, header_data):
        return header_data[0][1][1]

    def set_header_data(self, header_data):
        try:
            _form_header = self.get_decode_header(header_data)
            helper.decode_header(self.header, _form_header)
            self.header['data'] = header_data
        except Exception as err:
            raise ExtException(parent=err)

    def decode_includes(self, src_dir, dest_dir, dest_path, header_data):
        tasks = []
        includes = self.get_decode_includes(header_data)
        for include in includes:
            self.decode_include(src_dir, dest_dir, dest_path, tasks, include)
        return tasks

    def decode_include(self, src_dir, dest_dir, dest_path, tasks, include):
        try:
            count_include_types = int(include[2])
        except IndexError:
            raise ExtException(message='Include types not found', detail=self.__class__.__name__)
        for i in range(count_include_types):
            _metadata = include[i + 3]
            _count_obj = int(_metadata[1])
            _metadata_type_uuid = _metadata[0]
            try:
                metadata_type = MetaDataTypes(_metadata_type_uuid)
            except ValueError:
                # data = helper.json_read(src_dir, f'{_metadata[2]}.json')  # чтобы посмотреть что это
                # continue

                if not _count_obj:
                    continue

                if not isinstance(_metadata[2], str):  # вложенный объект
                    continue
                msg = f'У {self.__class__.__name__} {self.header["name"]} неизвестный тип вложенных метаданных: {_metadata_type_uuid} лежит в файле {_metadata[2]}'
                print(msg)
                continue
                # raise Exception(msg)
            if not _count_obj:
                continue
            new_dest_path = os.path.join(dest_path, metadata_type.name)
            external_obj = False
            for j in range(_count_obj):
                obj_uuid = _metadata[j + 2]
                if isinstance(obj_uuid, str):
                    if j == 0:
                        os.mkdir(os.path.join(dest_dir, new_dest_path))

                    tasks.append([metadata_type.name, [src_dir, obj_uuid, dest_dir, new_dest_path, self.options]])
                    external_obj = True
                elif isinstance(obj_uuid, list):
                    if not metadata_type:
                        continue
                    try:
                        handler = helper.get_class_metadata_object(metadata_type.name)
                    except Exception as err:
                        continue
                    if j == 0:
                        os.mkdir(os.path.join(dest_dir, new_dest_path))
                    handler.decode_local_include(self, obj_uuid, src_dir, dest_dir, new_dest_path, self.options)
                    external_obj = True
            if external_obj and self.get_options('auto_include'):  # todo dynamic index
                include[i + 3] = metadata_type.name

    @classmethod
    def get_decode_includes(cls, header_data: list) -> list:
        raise NotImplementedError(f'Для метаданных {cls.__name__} не указана ссылка на includes')

    def fill_header_includes(self, include_index):
        try:
            includes = self.get_decode_includes(self.header['data'])
        except NotImplementedError:
            return
        for include in includes:
            try:
                count_include_types = int(include[2])
            except IndexError:
                raise ExtException(message='Include types not found', detail=self.__class__.__name__)
            for i in range(count_include_types):
                _metadata = include[i + 3]
                if isinstance(_metadata, str):
                    metadata_type = MetaDataTypes[_metadata]
                    include_objects = include_index.get(_metadata, [])
                    include[i + 3] = [metadata_type.value, str(len(include_objects)), *include_objects]

    def encode_includes(self, src_dir, file_name, dest_dir, parent_id):
        tasks = []
        includes = []
        entries = sorted(os.listdir(src_dir))
        for entry in entries:
            src_entry_path = os.path.join(src_dir, entry)
            if os.path.isdir(src_entry_path):
                try:
                    MetaDataTypes[entry]
                except KeyError:
                    raise Exception(
                        f'Обнаружен не поддерживаемый тип метаданных или некорректно описанная область {entry} ')
                includes.append(entry)

        for include in includes:
            handler = helper.get_class_metadata_object(include)
            _src_dir = os.path.join(src_dir, include)
            handler.encode_get_include_obj(_src_dir, dest_dir, handler.get_obj_name(), tasks,
                                           self.options, parent_id, {})
        return tasks

    @classmethod
    def encode_get_include_obj(cls, src_dir, dest_dir, include, tasks, options, parent_id, include_index):
        """
        возвращает список задач на парсинг объектов этого типа
        """
        entries = os.listdir(src_dir)
        for entry in entries:
            if cls.re_meta_data_obj.fullmatch(entry):
                tasks.append([include, [src_dir, entry[:-5], dest_dir, options, parent_id, include_index]])

    @classmethod
    def encode_versions(cls, file_list):
        versions = ["1", str(len(file_list) + 1), helper.str_encode(""), str(uuid4())]
        for file_name in file_list:
            versions.append(helper.str_encode(file_name))
            versions.append(str(uuid4()))
        return [versions]

    @classmethod
    def encode_get_include_obj_from_named_folder(cls, src_dir, dest_dir, include, tasks, options, parent_id,
                                                 include_index):
        """
        возвращает список задач на парсинг объектов этого типа
        """
        entries = os.listdir(src_dir)
        for entry in entries:
            if os.path.isdir(os.path.join(src_dir, entry)):
                new_src_dir = os.path.join(src_dir, entry)
                tasks.append([include, [new_src_dir, entry, dest_dir, options, parent_id, include_index]])

    def encode_version(self):
        return self.header['version']

    @classmethod
    def get_class_name_without_version(cls, version=None):
        _version = version if version else cls.version
        if _version and cls.__name__.endswith(_version):
            return cls.__name__[:len(_version) * -1]
        return cls.__name__

    def read_raw_code(self, src_dir, file_name, encoding=None):
        code = helper.txt_read(src_dir, file_name, encoding=encoding)
        if code:
            # if self.version in ['801', '802']:  # убираем комментрии у директив
            code = self.directive_1c_comment.sub(r'\g<n>\g<d>', code)
        return code

    def write_raw_code(self, code, dest_dir, filename, encoding='uft-8'):
        if code is not None:
            if self.version in ['801', '802']:  # комментируем директивы
                code = self.directive_1c_uncomment.sub(r'\g<n>// v8unpack \g<d>', code)
            helper.txt_write(code, dest_dir, filename, encoding=encoding)

    def decode_code(self, src_dir):
        for code_name in self.ext_code:
            if code_name in self.code:
                continue  # код был в файле с формой
            _obj_code_dir = f'{os.path.join(src_dir, self.header["uuid"])}.{self.ext_code[code_name]}'
            if os.path.isdir(_obj_code_dir):
                self.header[f'code_info_{code_name}'] = helper.brace_file_read(_obj_code_dir, 'info')
                try:
                    self.code[code_name] = self.read_raw_code(_obj_code_dir, 'text')
                    encoding = helper.detect_by_bom(os.path.join(_obj_code_dir, 'text'), 'utf-8')
                    self.header[f'code_encoding_{code_name}'] = encoding  # можно безболезненно поменять на utf-8-sig
                except FileNotFoundError as err:
                    # todo могут быть зашифрованные модули тогда файл будет # image.json - зашифрованный контент
                    not_encrypted = True
                    for encrypted_type in self.encrypted_types:
                        if os.path.isfile(os.path.join(_obj_code_dir, encrypted_type)):
                            self.code[code_name] = helper.bin_read(_obj_code_dir, encrypted_type)
                            self.header[f'code_encoding_{code_name}'] = encrypted_type
                            not_encrypted = False
                            break
                    if not_encrypted:
                        raise err from err
            else:
                try:
                    code_file_name = f'{self.header["uuid"]}.{self.ext_code[code_name]}'
                    self.code[code_name] = self.read_raw_code(src_dir, code_file_name)
                    encoding = helper.detect_by_bom(os.path.join(src_dir, code_file_name), 'utf-8')
                    self.header[f'code_info_{code_name}'] = 'file'
                    self.header[f'code_encoding_{code_name}'] = encoding  # можно безболезненно поменять на utf-8-sig
                except FileNotFoundError as err:
                    pass

    def write_decode_code(self, dest_dir, file_name):
        for code_name in self.code:
            if self.code[code_name]:
                if self.header.get(f'code_encoding_{code_name}') in self.encrypted_types:
                    helper.bin_write(self.code[code_name], dest_dir, self.header[f'code_encoding_{code_name}'])
                else:
                    helper.txt_write(self.code[code_name], dest_dir, f'{file_name}.{code_name}.bsl')

    def encode_code(self, src_dir, file_name):
        for code_name in self.ext_code:
            if self.header.get(f'code_info_{code_name}'):
                try:
                    if self.header.get(f'code_encoding_{code_name}') in self.encrypted_types:
                        self.code[code_name] = helper.bin_read(src_dir, self.header.get(f'code_encoding_{code_name}'))
                    else:
                        self.code[code_name] = helper.txt_read(src_dir, f'{file_name}.{code_name}.bsl')
                except FileNotFoundError:
                    self.code[code_name] = ''

    def write_encode_code(self, dest_dir):
        for code_name in self.code:
            encoding = self.header.get(f'code_encoding_{code_name}', 'utf-8')
            if self.header[f'code_info_{code_name}'] == 'file':
                _code_file_name = f'{self.header["uuid"]}.{self.ext_code[code_name]}'
                self.file_list.append(_code_file_name)
                self.write_raw_code(self.code[code_name], dest_dir, _code_file_name, encoding=encoding)
            else:
                dir_name = f'{self.header["uuid"]}.{self.ext_code[code_name]}'
                _code_dir = os.path.join(dest_dir, dir_name)
                self.file_list.append(dir_name)
                helper.makedirs(_code_dir)
                helper.brace_file_write(self.header[f'code_info_{code_name}'], _code_dir, 'info')
                if encoding in self.encrypted_types:
                    helper.bin_write(self.code[code_name], _code_dir, encoding)
                else:
                    self.write_raw_code(self.code[code_name], _code_dir, 'text', encoding=encoding)

    def set_product_version(self, product_version):
        pass

    def set_product_comment(self, product_version):
        header = self.get_decode_header(self.header['data'])
        header[4] = helper.str_encode(helper.str_decode(header[4]) + product_version)

    def set_product_info(self, src_dir, file_name):
        product_version = ''
        try:
            product_version = helper.txt_read(src_dir, 'version.bin', encoding='utf-8')
            self.set_product_version(product_version)
        except FileNotFoundError:
            pass
        if file_name:
            product_info = f';{file_name};{product_version}'
            self.set_product_comment(product_info)

    def _decode_html_data(self, src_dir, dest_dir, dest_file_name, *, header_field='html', file_number=0,
                          extension='html'):
        try:
            data = helper.brace_file_read(src_dir, f'{self.header["uuid"]}.{file_number}')
        except FileNotFoundError:
            return
        try:
            if data[0][3] and data[0][3][0]:
                bin_data = self._extract_b64_data(data[0][3])
                helper.bin_write(bin_data, dest_dir, f'{dest_file_name}.{extension}')
        except IndexError:
            pass
        self.header[header_field] = data

    def _extract_b64_data(self, raw_data):
        if raw_data[0].startswith('##base64:'):
            bin_data = b64decode(raw_data[0][9:])
            raw_data[0] = '##base64:'
        elif raw_data[0].startswith('#base64:'):
            bin_data = b64decode(raw_data[0][8:])
            raw_data[0] = '#base64:'
        elif raw_data[0].startswith('#data:'):
            bin_data = b64decode(raw_data[0][6:])
            raw_data[0] = '#data:'
        else:
            raise NotImplementedError('decode_html_data')
        return bin_data

    def _encode_html_data(self, src_dir, file_name, dest_dir, *, header_field='html', file_number=0, extension='html'):
        try:
            bin_data = helper.bin_read(src_dir, f'{file_name}.{extension}')
        except FileNotFoundError:
            bin_data = None
        header = self.header.get(header_field)
        if header and len(header[0]) > 2 and bin_data:
            header[0][3][0] += b64encode(bin_data).decode(encoding='utf-8')
        if header:
            file_name = f'{self.header["uuid"]}.{file_number}'
            self.file_list.append(file_name)
            helper.brace_file_write(header, dest_dir, file_name)

    @staticmethod
    def _get_b64_string(bin_data):
        if not bin_data:
            return "##base64:"
        else:
            return "#base64:" + b64encode(bin_data).decode(encoding='utf-8')

    def _decode_info(self, src_dir, dest_dir, dest_file_name):
        if self._obj_info:
            for elem in self._obj_info:
                try:
                    data = helper.brace_file_read(src_dir, f'{self.header["uuid"]}.{self._obj_info[elem]}')
                    helper.json_write(data, dest_dir, f'{dest_file_name}.{self._obj_info[elem]}.json')

                except FileNotFoundError:
                    pass

    def _encode_info(self, src_dir, file_name, dest_dir):
        if self._obj_info:
            for elem in self._obj_info:
                try:
                    data = helper.json_read(src_dir, f'{file_name}.{self._obj_info[elem]}.json')
                    dest_file_name = f'{self.header["uuid"]}.{self._obj_info[elem]}'
                    helper.brace_file_write(data, dest_dir, dest_file_name)
                    self.file_list.append(dest_file_name)
                except FileNotFoundError:
                    pass



from ..MetaDataObject.core.Container import Container


class Catalog(Container):
    ext_code = {
        'mgr': '3',  # модуль менеджера Справочника
        'obj': '0',  # модуль объекта Справочника
    }
    help_file_number = 1
    predefined_file_number = '1c'

    @classmethod
    def get_decode_header(cls, header):
        return header[0][1][9][1]



from ..MetaDataObject.core.Container import Container


class Task(Container):
    ext_code = {
        'obj': 6,
        'mgr': 7
    }
    help_file_number = 5
    # @classmethod
    # def get_decode_header(cls, header):
    #     return header[0][1][9][1]



from ..MetaDataObject.core.Container import Container
from .. import helper


class BusinessProcess(Container):
    ext_code = {
        'obj': 6,
        'mgr': 8
    }
    help_file_number = 5
    pass

    def decode_object(self, src_dir, file_name, dest_dir, dest_path, version, header_data):
        super().decode_object(src_dir, file_name, dest_dir, dest_path, version, header_data)
        try:
            package = helper.bin_read(src_dir, f'{self.header["uuid"]}.7')
            helper.bin_write(package, self.new_dest_dir, 'Карта маршрута.bin')
        except FileNotFoundError:
            return

    def encode_object(self, src_dir, file_name, dest_dir):
        res = super().encode_object(src_dir, file_name, dest_dir)
        try:
            package = helper.bin_read(src_dir, 'Карта маршрута.bin')
            file_name = f'{self.header["uuid"]}.7'
            helper.bin_write(package, dest_dir, file_name)
            self.file_list.append(file_name)
        except FileNotFoundError:
            return
        return res


import sys
import unittest

from v8unpack import helper

sys.path.append("../../src/")


class TestHelper(unittest.TestCase):
    def test_get_extension_from_comment(self):
        data = [
            ("", "bin"),
            ("png", "png"),
            (" png ", "png"),
            ("Комментарий png", "png")
        ]
        for elem in data:
            self.assertEqual(elem[1], helper.get_extension_from_comment(elem[0]))

    def test_check_version(self):
        data = [
            ("12.1.1", "12.1.0", False),
            ("12.1.1", "12.1.1", False),
            ("12.1.1", "12.1.2", False),
            ("12.1.1", "13.1.1", True),
            ("12.1.1", "12.2.1", True),
            ("", "", True),
            ("F", "A", True),
            ("12.1.1", "", True),
            ("12.1.1", "d", True)
        ]
        for elem in data:
            try:
                helper.check_version(elem[0], elem[1])
                self.assertFalse(elem[2])
            except AssertionError:
                self.assertTrue(elem[2])



from enum import Enum

from .... import helper
from ....ext_exception import ExtException


def calc_offset(counters, raw_data):
    # counters - позиции указывающие на счетчики, если не 0 то за ним идет столько записей размера size
    index = 0
    for counter_index, size in counters:
        index += counter_index
        if size:
            value = int(raw_data[index])
            index += value * size
    return index


def check_count_element(counters, raw_data):
    # counters - позиции указывающие на счетчики, если не 0 то за ним идет столько записей размера size
    index = 0
    var_len = 0
    for counter_index, size in counters:
        index += counter_index
        if size:
            value = int(raw_data[index])
            index += value * size
            var_len += value * size
    return len(raw_data) - var_len


class FormItemTypes(Enum):
    Field = '77ffcc29-7f2d-4223-b22f-19666e7250ba'
    Button = 'a9f3b1ac-f51b-431e-b102-55a69acdecad'
    Decoration = '3d3cb80c-508b-41fa-8a18-680cdf5f1712'
    Group = 'cd5394d0-7dda-4b56-8927-93ccbe967a01'
    Table = '143c00f7-a42d-4cd7-9189-88e4467dc768'
    ItemAddition = 'c5259a1d-518a-4afd-b98d-0176027e4feb'


class FormElement:

    @classmethod
    def get_name_node_offset(cls, raw_data):
        return calc_offset([(3, 1), (1, 0)], raw_data)

    @classmethod
    def decode(cls, form, path, raw_data):
        offset = cls.get_name_node_offset(raw_data)
        name = raw_data[offset]
        if not isinstance(name, str) or not name:
            raise ExtException(message='form elem name not string')
        form.elements_data[f'{path}/{helper.str_decode(name)}'] = raw_data
        return dict(
            name=helper.str_decode(name),
            type=cls.__name__,
            # raw=raw_data
        )

    @classmethod
    def encode(cls, form, path, data):
        return form.elements_data[f"{path}/{data['name']}"]

    @classmethod
    def decode_list(cls, form, raw_data, index_element_count, path=''):
        try:
            result = []
            element_count = int(raw_data[index_element_count])
            if not element_count:
                return

            for i in range(element_count):
                metadata_type_uuid = raw_data[index_element_count + i * 2 + 1]
                elem_raw_data = raw_data[index_element_count + i * 2 + 2]
                try:
                    metadata_type = FormItemTypes(metadata_type_uuid)
                except ValueError:
                    raise ExtException(
                        message='Неизвестный тип элемента формы',
                        detail=f'{form.__class__.__name__} {form.header["name"]} : {metadata_type_uuid}'
                    )
                try:
                    handler = cls.get_class_form_elem(metadata_type.name)
                except Exception as err:
                    raise ExtException(
                        parent=err,
                        message='Проблема с парсером элемента формы',
                        detail=f'{metadata_type.name} - {err}'
                    )
                try:
                    elem_data = handler.decode(form, path, elem_raw_data)
                except helper.FuckingBrackets as err:
                    raise err from err
                except Exception as err:
                    raise ExtException(
                        parent=err,
                        detail=f'{metadata_type.name} - {err}',
                        message='Ошибка разбора элемента формы'
                    )
                result.append(elem_data)

            raw_data[index_element_count] = 'Дочерние элементы отдельно'
            del raw_data[index_element_count + 1:index_element_count + 1 + element_count * 2]
            return result
        except helper.FuckingBrackets as err:
            raise err
        except Exception as err:
            raise ExtException(parent=err)

    @staticmethod
    def get_class_form_elem(name):
        return helper.get_class(f'v8unpack.MetaDataObject.versions.Form803Elements.{name}.{name}')

    @classmethod
    def encode_list(cls, form, items, raw_data, index_element_count, path=''):
        result = []
        for item in items:
            try:
                metadata_type = FormItemTypes[item['type']]
            except ValueError:
                raise ExtException(
                    message='Неизвестный тип элемента формы',
                    detail=f'{form.__class__.__name__} {form.header["name"]} : {item["type"]}'
                )
            try:
                handler = cls.get_class_form_elem(metadata_type.name)
            except Exception as err:
                raise ExtException(
                    message='Проблема с парсером элемента формы',
                    detail=f'{metadata_type.name} - {err}'
                )
            elem_data = handler.encode(form, path, item)
            result.append(metadata_type.value)
            result.append(elem_data)
        raw_data[index_element_count] = str(len(items))
        raw_data[index_element_count + 1:index_element_count + 1] = result

        return result


class _FormRoot:
    element_node_offset = [[2, 1], [3, 1], [4, 1]]
    name = 0
    index = 0
    index_name = 0

    @classmethod
    def decode(cls, raw_data):
        return dict(
            name=helper.str_decode(raw_data[cls.index_name]),
            raw=raw_data
        )

    @classmethod
    def decode_list(cls, form, raw_data):
        try:
            if len(raw_data) <= cls.index:
                return
            items = raw_data[cls.index]
            index_element_count = 1
            element_count = int(items[index_element_count])
            if not element_count:
                return

            result = []
            for i in range(element_count):
                result.append(cls.decode(items[i + index_element_count + 1]))
            items[index_element_count] = 'Дочерние элементы отдельно'
            del items[index_element_count + 1:index_element_count + 1 + element_count]
            return result
        except Exception as err:
            pass

    @classmethod
    def encode_list(cls, form, src_dir, file_name, version):
        raw_data = form.form[0][0]

        if len(raw_data) <= cls.index:
            return
        raw_data = raw_data[cls.index]
        result = []
        index_element_count = 1
        if raw_data[index_element_count] == 'Дочерние элементы отдельно':
            items = helper.json_read(src_dir, f'{file_name}.{cls.name}{version}.json')
            if items:
                for item in items:
                    result.append(cls.encode(form, item))
                raw_data[index_element_count] = str(len(items))
                raw_data[index_element_count + 1:index_element_count + 1] = result
            else:
                raw_data[index_element_count] = '0'

    @classmethod
    def encode(cls, form, data):
        return data['raw']


class FormParams(_FormRoot):
    name = 'params'
    index = 4
    index_name = 1


class FormProps(_FormRoot):
    name = 'props'
    index = 3
    index_name = 3


class FormCommands(_FormRoot):
    name = 'commands'
    index = 5
    index_name = 2



from ..MetaDataObject.core.Container import Container
from .. import helper


class ExchangePlan(Container):
    ext_code = {'obj': 2, 'mgr': 3}
    help_file_number = 0
    pass

    @classmethod
    def get_decode_header(cls, header):
        return header[0][1][12]

    def decode_object(self, src_dir, file_name, dest_dir, dest_path, version, header_data):
        super().decode_object(src_dir, file_name, dest_dir, dest_path, version, header_data)
        try:
            self.header['info'] = helper.brace_file_read(src_dir, f'{self.header["uuid"]}.1')
        except FileNotFoundError:
            self.header['info'] = None

    def write_encode_object(self, dest_dir):
        super().write_encode_object(dest_dir)
        if self.header['info']:
            file_name = f'{self.header["uuid"]}.1'
            helper.brace_file_write(self.header['info'], dest_dir, file_name)
            self.file_list.append(file_name)



from .Form803 import Form803


class ReportForm803(Form803):
    def get_decode_obj_header(self, header):
        return header[0][1][1]

    # @classmethod
    # def get_decode_header(cls, header_data):
    #     return header_data[0][1][1][1][1]

    # def encode_header(self):
    #     return [[
    #         "1",
    #         [
    #             "1",
    #             [
    #                 "0",
    #                 self.encode_header_title(),
    #             ],
    #             "0"
    #         ]]]



from ..MetaDataObject.core.Container import Container


class InformationRegister(Container):
    ext_code = {
        'mgr': '2',  # модуль менеджера
        'obj': '1',  # Модуль набора записей
    }
    help_file_number = 0

    @classmethod
    def get_decode_header(cls, header_data):
        return header_data[0][1][15][1]



from ..MetaDataObject import MetaDataObject
from ..MetaDataObject.versions.CommonTemplate803 import CommonTemplate803


class CommonTemplate(MetaDataObject):
    versions = {
        '803': CommonTemplate803,
        4: CommonTemplate803
    }



import os
import sys
import unittest

sys.path.append("../../src/")
from v8unpack import helper
from v8unpack.MetaDataObject.versions.Form803 import Form803 as Form
from v8unpack.MetaDataObject.CatalogForm import CatalogForm
from v8unpack.unittest_helper import compare_file, NotEqualLine


class TestDecodeBraceFile(unittest.TestCase):
    def setUp(self) -> None:
        self.current_dir = os.path.dirname(__file__)
        self.data_dir = os.path.join(self.current_dir, 'data', 'form')
        self.temp_dir = os.path.join(self.data_dir, 'temp')
        self.temp_dir2 = os.path.join(self.data_dir, 'temp2')

    def test_decode_all_raw_data(self):
        test_files = os.listdir(self.data_dir)
        result = ''
        for file_name in test_files:
            if file_name == 'temp':
                continue
            uuid, x = file_name.split('.')
            result += self.decode_raw_data(uuid)
        self.assertEqual('', result)

    def test_decode_file(self):
        uuid = '01f67911-4ca7-4668-b970-62c299b98e56'
        result = self.decode_raw_data(uuid)
        self.assertEqual('', result)

    def decode_raw_data(self, file_name):
        helper.clear_dir(self.temp_dir)
        helper.clear_dir(self.temp_dir2)
        options = dict(version='803')
        # helper.clear_dir(self.temp_dir2)
        CatalogForm.decode(self.data_dir, file_name, self.temp_dir, '', options=options)
        temp_dir = os.path.join(self.temp_dir, 'ФормаСписка')
        form = Form(options=options)
        form.encode(temp_dir, file_name, self.temp_dir2, None, [{}])
        problems = ''
        try:
            result = compare_file(
                os.path.join(self.data_dir, file_name),
                os.path.join(self.temp_dir2, file_name),
                problems
            )
        except NotEqualLine as err:
            result = err
        return result

    def test_decode_by_scheme(self):
        from v8unpack.MetaDataObject.schemes.Form import Form

        file_name = '9c80a4da-8f1f-48ae-915c-4227fef8f364'
        data = Form(self.data_dir, file_name).decode()



from ..MetaDataObject.core.Container import Container


class Document(Container):
    help_file_number = 1
    ext_code = {
        'obj': 0,
        'mgr': 2
    }

    @classmethod
    def get_decode_header(cls, header):
        return header[0][1][9][1]



import os
from base64 import b64decode, b64encode

from .. import helper
from ..MetaDataObject.core.Simple import Simple
from ..ext_exception import ExtException


class CommonPicture(Simple):
    def __init__(self, *, obj_name=None, options=None):
        super().__init__(obj_name=obj_name, options=options)
        self.ext_code = {}
        self.data = None
        self.raw_data = None

    def decode_object(self, src_dir, file_name, dest_dir, dest_path, version, header_data):
        try:
            super().decode_object(src_dir, file_name, dest_dir, dest_path, version, header_data)
            try:
                self.header['info'] = helper.brace_file_read(src_dir, f'{self.header["uuid"]}.0')
            except FileNotFoundError:
                return
            if self.header['info'][0][2] and self.header['info'][0][2][0] and self.header['info'][0][2][0][0]:
                bin_data = self._extract_b64_data(self.header['info'][0][2][0])

                extension = helper.get_extension_from_comment(self.header['comment'])
                _dest_dir = os.path.join(dest_dir, dest_path)
                if dest_dir:
                    helper.bin_write(bin_data, _dest_dir, f'{self.header["name"]}.{extension}')
        except Exception as err:
            raise ExtException(parent=err)

    def encode_object(self, src_dir, file_name, dest_dir):
        super().encode_object(src_dir, file_name, dest_dir)

        extension = helper.get_extension_from_comment(self.header['comment'])
        try:
            bin_data = helper.bin_read(src_dir, f'{self.header["name"]}.{extension}')
            self.header['info'][0][2][0][0] += b64encode(bin_data).decode(encoding='utf-8')
            file_name = f'{self.header["uuid"]}.0'
            helper.brace_file_write(self.header['info'], dest_dir, file_name)
            self.file_list.append(file_name)
        except FileNotFoundError:
            pass



# -*- coding: utf-8 -*-
import collections
import datetime
import io
import math
import os
import tempfile
import zlib
from datetime import datetime, timedelta
from struct import pack
# -*- coding: utf-8 -*-
from struct import unpack

from .helper import file_size

Header = collections.namedtuple('Header', 'first_empty_block_offset, default_block_size, count_files')
Block = collections.namedtuple('Block', 'doc_size, current_block_size, next_block_offset, data')
File = collections.namedtuple('File', 'name, size, created, modified, data')
DocumentData = collections.namedtuple('DocumentData', 'size, data')

# Размер буффера передачи данных из потока в поток
BUFFER_CHUNK_SIZE = 512


class Document:
    def __init__(self, container):
        self.container = container
        self.full_size = 0  # include all header size
        self.data_size = 0

    def read(self, file, offset, min_block_size=0):
        document_data = self.read_chunk(file, offset, min_block_size)
        return b''.join([chunk for chunk in document_data])

    def read_chunk(self, file, offset, min_block_size=0):
        """
        Считывает документ из контейнера. В качестве данных документа возвращается генератор.

        :param file: объект файла контейнера
        :type file: BufferedReader
        :param offset: смещение документа в контейнере
        :type offset: int
        :return: данные документа
        :rtype:
        """
        gen = self._read_gen(file, offset, min_block_size)

        try:
            self.data_size = next(gen)
        except StopIteration:
            self.data_size = 0

        return gen

    def _read_gen(self, file, offset, min_block_size=0):
        """
        Создает генератор чтения данных документа в контейнере.
        Первое значение генератора - размер документа (байт).
        Остальные значения - данные блоков, составляющих документ

        :param file: объект файла контейнера
        :type file: BufferedReader
        :param offset: смещение документа в контейнере (байт)
        :type offset: int
        :return: генератор чтения данных документа
        """
        header_block = self.read_block(file, offset)
        doc_size = max(header_block.doc_size, min_block_size)
        if doc_size > header_block.current_block_size:
            self.full_size = doc_size + math.ceil(
                header_block.doc_size / header_block.current_block_size) * min_block_size
            if min_block_size == self.container.index_block_size:
                self.full_size += self.container.index_block_size
        else:
            self.full_size = doc_size + self.container.block_header_size

        if header_block is None:
            return
        else:
            yield header_block.doc_size
            yield header_block.data

            left_bytes = header_block.doc_size - len(header_block.data)
            next_block_offset = header_block.next_block_offset

            while left_bytes > 0 and next_block_offset != self.container.end_marker:
                block = self.read_block(file, next_block_offset, left_bytes)
                left_bytes -= len(block.data)
                yield block.data
                next_block_offset = block.next_block_offset

    def read_block(self, file, offset, max_data_length=None):
        """
        Считывает блок данных из контейнера.

        :param file: объект файла контейнера
        :type file: BufferedReader
        :param offset: смещение блока в файле контейнера (байт)
        :type offset: int
        :param max_data_length: максимальный размер считываемых данных из блока (байт)
        :type max_data_length: int
        :return: объект блока данных
        :rtype: Block
        """
        file.seek(offset + self.container.offset)
        header_size = self.container.block_header_size
        buff = file.read(header_size)
        if not buff:
            return
        header = unpack(self.container.block_header_fmt, buff)

        doc_size = int(header[1], 16)
        current_block_size = int(header[3], 16)
        next_block_offset = int(header[5], 16)

        if max_data_length is None:
            max_data_length = min(current_block_size, doc_size)

        data_size = min(current_block_size, max_data_length)

        data = file.read(data_size)

        return Block(doc_size, current_block_size, next_block_offset, data)

    def write_header(self, file, file_name):
        modify_time = epoch2int(os.stat(file.fileno()).st_mtime)
        # В *nix это не время создания файла.
        creation_time = epoch2int(os.stat(file.fileno()).st_ctime)
        buffer = b''.join([pack('QQi', creation_time, modify_time, 0), file_name.encode('utf-16-le'), b'\x00' * 4])
        attribute_doc_offset = self.write(io.BytesIO(buffer))
        return attribute_doc_offset

    def write(self, data, *, min_block_size=0):
        return self.write_block(data, doc_size=file_size(data), min_block_size=min_block_size)

    def write_block(self, data, *, doc_size=0, min_block_size=0, offset=None, next_block_offset=None):
        """
        Записывает блок данных в контейнер

        :param data: file-like объект
        :param kwargs: Опциональные параметры
        :return: смещение записанных данных (байт)
        :rtype: int
        """
        if offset is not None:
            self.container.file.seek(offset)
        else:
            offset = file_size(self.container.file)
        block_size = file_size(data)
        min_block_size = max(min_block_size, block_size)
        if not next_block_offset:
            next_block_offset = self.container.end_marker

        header_data = ('\r\n', self.container.int2hex(doc_size), ' ', self.container.int2hex(min_block_size), ' ',
                       self.container.int2hex(next_block_offset), ' ', '\r\n')
        header = pack(self.container.block_header_fmt, *[x.encode() for x in header_data])

        self.container.file.write(header)
        self.write_block_data(data, self.container.file)

        self.container.file.write(b'\x00' * (min_block_size - data.tell()))

        return offset

    @classmethod
    def compress(cls, src_fd, dest_fd):
        with tempfile.TemporaryFile() as f:
            compressor = zlib.compressobj(wbits=-15)
            src_fd.seek(0)
            while True:
                chunk = src_fd.read(BUFFER_CHUNK_SIZE)
                if not chunk:
                    f.write(compressor.flush())
                    break
                f.write(compressor.compress(chunk))
            cls.write_block_data(f, dest_fd)
        # return data_doc_offset

    @staticmethod
    def write_block_data(data, dest_file):
        data.seek(0)
        while True:
            buffer = data.read(BUFFER_CHUNK_SIZE)
            if not buffer:
                break
            dest_file.write(buffer)


def epoch2int(epoch_time):
    """
    Преобразует время в формате "количество секунд с начала эпохи" в количество сотых микросекундных интервалов
    с 0001.01.01

    :param epoch_time: время в формате Python
    :type epoch_time: real
    :return: количество сотых микросекундных интервалов
    :rtype: int
    """
    # Начало эпохи на разных системах - разная дата
    # Поэтому явно вычисляем разницу между указанной датой и 0001.01.01
    return (datetime.fromtimestamp(epoch_time) - datetime(1, 1, 1)) // timedelta(
        microseconds=100)



import json
import os
import sys
import unittest

from tqdm.auto import tqdm

from . import helper
from .container_reader import extract, decompress_and_extract
from .container_writer import build, compress_and_build
from .decoder import decode, encode
from .file_organizer import FileOrganizer
from .file_organizer_ce import FileOrganizerCE
from .json_container_decoder import JsonContainerDecoder


class HelperTestDecode(unittest.TestCase):
    pool = None
    processes = None

    def tearDown(self) -> None:
        helper.close_pool(self.pool)

    def setUp(self) -> None:
        if not sys.warnoptions:
            import warnings
            warnings.simplefilter("ignore")

        self.pool = helper.get_pool(processes=self.processes)
        self.src_dir = os.path.join(sys.path[0])  # абсолютный путь до папки с исходным файлом
        self.src_file = ''  # имя исходного файла
        self.test_dir = ''  # абсолютный путь до временной папки c файлами промежуточных стадий
        self.dest_dir = ''
        self.options = None
        self.result = None
        self.version = '803'
        self.index = None

    def init(self, *, index_file_name=None, **kwargs):
        if not self.test_dir:
            self.test_dir = os.path.join(self.src_dir, 'tmp')
        self.options = kwargs
        helper.makedirs(self.test_dir, exist_ok=True)

        self.decode_dir_stage0 = self.get_decode_folder(0)
        self.decode_dir_stage1 = self.get_decode_folder(1)
        self.decode_dir_stage2 = self.get_decode_folder(2)
        self.decode_dir_stage3 = self.get_decode_folder(3)
        self.decode_dir_stage4 = self.dest_dir if self.dest_dir else self.get_decode_folder(4)

        self.encode_dir_stage0 = self.get_encode_folder(0)
        self.encode_dir_stage1 = self.get_encode_folder(1)
        self.encode_dir_stage2 = self.get_encode_folder(2)
        self.encode_dir_stage3 = self.get_encode_folder(3)

        if index_file_name:
            self.index = helper.check_index(index_file_name)
        pass

    def get_decode_folder(self, stage):
        return os.path.join(self.test_dir, f'decode_stage_{stage}')

    def get_encode_folder(self, stage):
        return os.path.join(self.test_dir, f'encode_stage_{stage}')

    def decode_stage0(self):
        extract(os.path.join(self.src_dir, self.src_file), self.decode_dir_stage0, False, False)
        if self.result:
            files = os.listdir(self.decode_dir_stage0)
            self.assertEqual(len(files), self.result['count_root_files_stage1'], 'count_root_files_stage1')

    def decode_stage1(self):
        decompress_and_extract(self.decode_dir_stage0, self.decode_dir_stage1, pool=self.pool)
        if self.result:
            files = os.listdir(self.decode_dir_stage1)
            self.assertEqual(len(files), self.result['count_root_files_stage1'], 'count_root_files_stage1')

    def decode_stage2(self):
        pass
        # json_decode(self.decode_dir_stage1, self.decode_dir_stage2, pool=self.pool)
        # if self.result:
        #     files = os.listdir(self.decode_dir_stage2)
        #     self.assertEqual(len(files), self.result['count_root_files_stage1'], 'count_root_files_stage1')

    def decode_stage3(self):
        decode(self.decode_dir_stage1, self.decode_dir_stage3, pool=self.pool, options=self.options)
        if self.result:
            files = os.listdir(self.decode_dir_stage3)
            self.assertEqual(len(files), self.result['count_root_files_stage3'], 'count_root_files_stage3')

    def decode_stage4(self, descent=None):
        helper.clear_dir(os.path.normpath(self.decode_dir_stage4))
        if not self.index:
            return
        if descent:
            FileOrganizerCE.unpack(self.decode_dir_stage3, self.decode_dir_stage4,
                                   pool=self.pool, index=self.index, descent=descent)
        else:
            FileOrganizer.unpack(self.decode_dir_stage3, self.decode_dir_stage4, pool=self.pool, index=self.index)
        if self.result:
            files = os.listdir(self.decode_dir_stage4)
            self.assertEqual(len(files), self.result['count_root_files_stage4'], 'count_root_files_stage4')
            pass

    def encode_stage4(self, descent=None):
        helper.clear_dir(os.path.normpath(self.encode_dir_stage3))
        if not self.index:
            return
        if descent:
            FileOrganizerCE.pack(self.decode_dir_stage4, self.encode_dir_stage3,
                                 pool=self.pool, index=self.index, descent=descent)
        else:
            FileOrganizer.pack(self.decode_dir_stage4, self.encode_dir_stage3, pool=self.pool, index=self.index)
        if self.result:
            files = os.listdir(self.encode_dir_stage3)
            self.assertEqual(len(files), self.result['count_root_files_stage3'], 'count_root_files_stage3')

    def encode_stage3(self):
        encode(self.decode_dir_stage3, self.encode_dir_stage1, pool=self.pool,
               file_name=os.path.basename(self.src_file), options=self.options)
        self.assert_stage(self.decode_dir_stage1, self.encode_dir_stage1)
        if self.result:
            files = os.listdir(self.encode_dir_stage1)
            self.assertEqual(len(files), self.result['count_root_files_stage1'])

    def encode_stage2(self):
        pass
        # json_encode(self.encode_dir_stage2, self.encode_dir_stage1, pool=self.pool)
        # self.assert_stage(self.decode_dir_stage1, self.encode_dir_stage1)
        # if self.result:
        #     files = os.listdir(self.encode_dir_stage1)
        #     self.assertEqual(len(files), self.result['count_root_files_stage1'])

    def encode_stage1(self):
        compress_and_build(self.encode_dir_stage1, self.encode_dir_stage0, pool=self.pool)
        # self.assert_stage(self.decode_dir_stage0, self.encode_dir_stage0)
        if self.result:
            files = os.listdir(self.encode_dir_stage0)
            self.assertEqual(len(files), self.result['count_root_files_stage1'])

    def encode_stage0(self):
        # helper.clear_dir(os.path.normpath(self.test_dir))
        encode_file_path = os.path.join(self.test_dir, self.src_file)
        # decode_file_path = os.path.join(self.src_dir, self.src_file)
        build(self.encode_dir_stage0, encode_file_path, True)
        # self.assertByteFile(decode_file_path, encode_file_path)

    def assert_external_data_processor_decode_stage3(self):
        if self.result:
            form_files = os.listdir(os.path.join(self.decode_dir_stage3, 'Form'))
            self.assertEqual(len(form_files), self.result['count_forms_files'])
            template_files = os.listdir(os.path.join(self.decode_dir_stage3, 'Template'))
            self.assertEqual(len(template_files), self.result['count_templates_files'])

    def assert_stage(self, decode_dir, encode_dir):
        problems = []
        self._assert_stage(decode_dir, encode_dir, problems, True)
        if problems:
            with open(f'{encode_dir}.txt', 'w', encoding='utf-8') as log:
                for problem in problems:
                    log.write(problem)
        self.assertTrue(not problems, f'files not equal - log in {encode_dir}.txt')

    def _assert_stage(self, decode_dir, encode_dir, problems, root=False):
        entries = os.listdir(decode_dir)
        include_problems = ''
        pbar = None
        if root:
            pbar = tqdm(desc=f'{"Сравниваем с оригиналом":30}', total=len(entries))
        for entry in entries:
            if pbar:
                pbar.update()
            if entry.startswith('configinfo'):
                continue
            path_decode_entry = os.path.join(decode_dir, entry)
            path_encode_entry = os.path.join(encode_dir, entry)
            if os.path.isdir(path_decode_entry):
                self._assert_stage(path_decode_entry, path_encode_entry, problems)
            else:
                try:
                    if entry == 'versions':
                        problem = compare_versions(decode_dir, encode_dir, problems)
                    else:
                        problem = compare_file(path_decode_entry, path_encode_entry, problems)
                except NotEqualLine as err:
                    problem = str(err)
                if problem:
                    problems.append(problem)
        # if problems:
        #     problems = f'   {decode_dir}\n   {encode_dir}{problems}\n'
        # if include_problems:
        #     problems += '\n' + include_problems
        if pbar:
            pbar.close()
        # return pr

    @classmethod
    def remove_closing_brackets(cls, file_data):
        return file_data

    def tmpl_json_encode(self, *, encode_src_dir='', entity_name='', encode_dest_dir='', decode_dir=''):
        JsonContainerDecoder.encode(encode_src_dir, f'{entity_name}.json', encode_dest_dir)
        if decode_dir:
            self.assertUtfFile(
                os.path.join(encode_dest_dir, entity_name),
                os.path.join(decode_dir, entity_name)
            )

    def assertUtfFile(self, src, dest):
        with open(src, 'r', encoding='utf-8') as file:
            src_data = file.read()
        with open(dest, 'r', encoding='utf-8') as file:
            dest_data = file.read()
        self.assertEqual(dest_data, src_data)

    def assertByteFile(self, src, dest):
        with open(src, 'rb') as file:
            src_data = file.read()
        with open(dest, 'rb') as file:
            dest_data = file.read()
        self.assertEqual(dest_data, src_data)


class NotEqualLine(Exception):
    pass


def compare_file(path_decode_entry, path_encode_entry, problems):
    def ignore():
        len_decode_line = len(decode_line)
        len_encode_line = len(encode_line)
        if decode_line.endswith(b'\r\n'):
            if decode_line[:-2] == encode_line:  # Лишние фигурные скобки
                return True
        if not encode_line:
            if decode_line == b'}\r\n' or decode_line == b'}':  # Лишние фигурные скобки
                return True
        if encode_line.endswith(b'\r\r\n'):
            if decode_line.endswith(b'\r\n') and decode_line[:-2] == encode_line[:-3]:
                decode_file.readline()
                return True
        if encode_line.endswith(b'\r\n'):
            if decode_line.endswith(b'\r\r\n') and decode_line[:-3] == encode_line[:-2]:
                encode_file.readline()
                # problems += f'\n      {entry:38} {i:6} \\r {decode_line}'
                return True
            if decode_line.endswith(b'\n') and decode_line[:-1] == encode_line[:-2]:
                return True
            if encode_line[:-2] == decode_line:
                return True
            # if decode_line.endswith(b'\r\n') \
            #         and len_decode_line > 2 \
            #         and decode_line[-3] == b','[0] \
            #         and (len_decode_line - len_encode_line) == 1:
            #     return True
        if encode_line.startswith(b'#base64') and encode_line[8:] == decode_line[9:]:
            return True
        if len_decode_line > 36 and len_decode_line == len_encode_line:  # допущение, т.к. в списоке инклюдов порядок теперь разный
            return True
        return False

    problems = ''
    entry = os.path.basename(path_encode_entry)
    folder = os.path.basename(os.path.dirname(path_encode_entry))
    title = os.path.join(folder, entry)

    with open(path_decode_entry, 'rb') as decode_file:
        try:
            i = 0
            with open(path_encode_entry, 'rb') as encode_file:
                for decode_line in decode_file:
                    # if i == 1 and decode_line.startswith(b'\xef\xbb\xbf'):
                    #     decode_line = decode_line[3:]
                    i += 1
                    encode_line = encode_file.readline()
                    if decode_line != encode_line:
                        if ignore():
                            continue
                        problems += f'\n      {title:73} {i:5} {encode_line[:30]}'
                        problems += f'\n      {"":79} {decode_line[:30]}'
                        raise NotEqualLine(problems)
                    elif decode_line.hex() != encode_line.hex():
                        a = 1
                for encode_line in encode_file:
                    i += 1
                    if encode_line == b'}\r\n' or b'}':
                        continue
                    problems += f'\n      {title:73} {i:5} {encode_line[:30]}'
                    raise NotEqualLine(problems)
        except FileNotFoundError:
            problems += f'\n      {title:73}  not encode file'
    return problems


def compare_versions(decode_dir, encode_dir, problems):
    def create_index(versions: list):
        index = []
        count = int((len(versions[0]) - 4) / 2)
        for i in range(count):
            j = i * 2 + 4
            elem = versions[0][j]
            index.append(elem)
        return index

    problems = ''
    decode_data = helper.brace_file_read(decode_dir, 'versions')
    decode_data = create_index(decode_data)
    encode_data = helper.brace_file_read(encode_dir, 'versions')
    encode_data = create_index(encode_data)
    size = len(decode_data)
    for i in range(size):
        j = size - i - 1
        try:
            elem = decode_data[j]
            encode_data.remove(elem)
            decode_data.pop(j)
        except ValueError:
            pass
    if encode_data or decode_data:
        problems += f'\n      {"versions":73} {"":5} {json.dumps(encode_data)}'
        problems += f'\n      {"":79} {json.dumps(decode_data)}'
    return problems



from ..MetaDataObject.core.Container import Container


class ChartOfCharacteristicType(Container):
    ext_code = {
        'mgr': '16',  # модуль менеджера Плана видов характеристик
        'obj': '15',  # модуль менеджера Плана видов характеристик
    }
    help_file_number = 5
    predefined_file_number = 7

    @classmethod
    def get_decode_header(cls, header_data):
        return header_data[0][1][13][1]



# -*- coding: utf-8 -*-
import collections
import datetime
import io
import math
import os
import tempfile
import zlib
from datetime import datetime, timedelta
from struct import pack, calcsize
# -*- coding: utf-8 -*-
from struct import unpack

from tqdm import tqdm

from .container_doc import Document
from .helper import clear_dir, file_size

Header = collections.namedtuple('Header', 'first_empty_block_offset, default_block_size, count_files')
Block = collections.namedtuple('Block', 'doc_size, current_block_size, next_block_offset, data')
File = collections.namedtuple('File', 'name, size, created, modified, data')
DocumentData = collections.namedtuple('DocumentData', 'size, data')

# Размер буффера передачи данных из потока в поток
BUFFER_CHUNK_SIZE = 512


class Container:
    end_marker = 0x7fffffff
    header_fmt = '4i'
    header_size = 16
    block_header_fmt = '2s8s1s8s1s8s1s2s'
    block_header_size = 31
    block_header_fmt_size = 8
    index_fmt = 'i'
    default_block_size = 0x200
    index_block_size = 0x200

    def __init__(self):
        self.file = None
        self.offset = 0
        self.first_empty_block_offset = None
        self.default_block_size = 0x200
        self.files = None
        self.size = 0
        self.toc = []

    def read(self, file, offset=0):
        self.offset = offset
        try:
            header = self.read_header(file)
        except Exception as err:
            raise err from err

        if header.default_block_size == 0:
            raise BufferError('Container is empty')

        self.file = file
        self.first_empty_block_offset = header.first_empty_block_offset
        self.default_block_size = header.default_block_size
        #: Список файлов в контейнере
        self.files = self.read_files(self.file)

    def extract(self, dest_dir, deflate=False, recursive=False, *, progress=None):
        """
        Распаковывает содержимое контейнера в каталог

        :param dest_dir: каталог распаковки
        :type dest_dir: string
        :param deflate: разархивировать содержимое файлов
        :type deflate: bool
        :param recursive: выполнять рекурсивно
        :type recursive: bool
        """

        clear_dir(dest_dir)
        if not self.files:
            print('Пустой контейнер = распаковывать нечего')
            return

        progress_bar = None
        if progress is not None:
            msg = f"Распаковываем контейнер {progress}"
            progress_bar = tqdm(desc=f'{msg:30}', total=len(self.files))

        for filename, file_obj in self.files.items():
            self.extract_file(filename, file_obj, dest_dir, deflate, recursive)
            if progress_bar:
                progress_bar.update()
        if progress_bar:
            progress_bar.close()

    @staticmethod
    def extract_file(filename, file_obj, path, deflate=False, recursive=False):
        file_path = os.path.join(path, filename)
        with open(file_path, 'wb') as f:
            if deflate:
                # wbits = -15 т.к. у архивированных файлов нет заголовков
                decompressor = zlib.decompressobj(-15)
                for chunk in file_obj.data:
                    decomressed_chunk = decompressor.decompress(chunk)
                    f.write(decomressed_chunk)
            else:
                for chunk in file_obj.data:
                    f.write(chunk)

        if not recursive:
            return

        # Каждый файл внутри контейнера может быть контейнером
        # Для проверки является ли файл контейнером проверим первые 4 бита
        # Способ проверки ненадежный - нужно придумать что-то другое
        file_is_container = False
        with open(file_path, 'rb') as f:
            if f.read(4) == b'\xFF\xFF\xFF\x7F':
                file_is_container = True
        if file_is_container:
            temp_name = file_path + '.tmp'
            os.rename(file_path, temp_name)
            with open(temp_name, 'rb') as f:
                _container = Container()
                _container.read(f)
                _container.extract(file_path, recursive=True)
            os.remove(temp_name)

    def read_header(self, file):
        """
        Считывыет заголовок контейнера.

        :param file: объект файла контейнера
        :type file: BufferedReader
        :return: Заголовок контейнера
        :rtype: Header
        """
        file.seek(0 + self.offset)
        buff = file.read(calcsize(self.header_fmt))
        header = unpack(self.header_fmt, buff)
        # if header[0] != self.end_marker:
        #     raise Exception('Bad container format')
        return Header(header[0], header[1], header[2])

    @classmethod
    def parse_datetime(cls, time):
        """
        Преобразует внутренний формат хранения дат файлов в контейнере в обычную дату

        :param time: внутреннее представление даты
        :type time: string
        :return: дата/время
        :rtype: datetime
        """
        # TODO проверить работу на *nix, т.к там начало эпохи - другая дата
        return datetime(1, 1, 1) + timedelta(microseconds=(time * 100))

    def read_files(self, file):
        """
        Считывает оглавление контейнера

        :param file: объект файла контейнера
        :type file: BufferedReader
        :return: словарь файлов в контейнере
        :rtype: OrderedDict
        """
        size = 0
        # Первый документ после заголовка содержит оглавление
        doc = Document(self)
        doc_data = doc.read(file, self.header_size, self.index_block_size)
        table_of_contents = [unpack(f'2{self.index_fmt}', x) for x in
                             doc_data.split(pack(self.index_fmt, self.end_marker))[:-1]]
        self.size += self.header_size + doc.full_size

        files = collections.OrderedDict()
        for file_description_offset, file_data_offset in table_of_contents:
            doc = Document(self)
            doc_data = doc.read(file, file_description_offset)
            self.size += doc.full_size
            fmt = ''.join(['QQi', str(doc.data_size - calcsize('QQi')), 's'])
            file_description = unpack(fmt, doc_data)
            # Из описания формата длина имени файла определяется точно, поэтому, теоретически, мусора быть не должно
            # По факту имя часто имеет в конце мусор, который чаще всего состоит из последовательности \x00 n-раз,
            # но иногда бывают и другие символы после \x00. Поэтому применяем вот такой костыль:
            name = file_description[3].decode('utf-16').partition('\x00')[0]

            doc = Document(self)
            file_data = doc.read_chunk(file, file_data_offset, self.default_block_size)
            self.size += doc.full_size

            inner_file = File(name, doc.data_size, self.parse_datetime(file_description[0]),
                              self.parse_datetime(file_description[1]), file_data)

            files[inner_file.name] = inner_file
        return files

    def build(self, file, src_dir, nested=False, *, offset=0):
        self.offset = offset
        self.file = file
        files = sorted(os.listdir(src_dir))
        self.write_header(len(files))
        for file_name in files:
            file_path = os.path.join(src_dir, file_name)
            if os.path.isdir(file_path):
                with tempfile.TemporaryFile() as tmp2:
                    _container = Container()
                    _container.build(tmp2, file_path, nested=True)
                self.add_file(tmp2, file_name, inflate=not nested)
            else:
                with open(file_path, 'rb') as entry_file:
                    self.add_file(entry_file, file_name, inflate=not nested)
        self.write_table_off_content()

    def write_header(self, count_files=0):
        """
        Записывает заголовок контейнера

        """
        self.file.seek(self.offset)
        self.file.write(
            pack(self.header_fmt, self.end_marker, self.default_block_size,
                 count_files, 0))
        # готовим место для оглавления
        self.file.write(b'\x00' * (self.index_block_size + self.block_header_size))

    def write_table_off_content(self):
        if len(self.toc) == 0:
            raise IOError('Container is empty')

        with tempfile.TemporaryFile() as f:
            for attr_offset, data_offset in self.toc:
                f.write(pack(f'3{self.index_fmt}', attr_offset, data_offset, self.end_marker))

            doc_size = file_size(f)

            total_blocks = math.ceil(doc_size / self.index_block_size)

            doc = Document(self)

            if total_blocks == 1:
                doc.write_block(f, doc_size=doc_size, offset=self.header_size)
            else:
                f.seek(0)
                next_block_offset = file_size(self.file)
                doc.write_block(io.BytesIO(f.read(self.index_block_size)),
                                doc_size=doc_size, offset=self.header_size,
                                min_block_size=self.index_block_size,
                                next_block_offset=next_block_offset)
                self.file.seek(0, os.SEEK_END)
                for i in range(1, total_blocks):
                    # 31 - длина заголовка блока
                    next_block_offset += self.index_block_size + self.block_header_size
                    doc.write_block(io.BytesIO(f.read(self.index_block_size)), doc_size=0,
                                    next_block_offset=next_block_offset, min_block_size=self.index_block_size)
                doc.write_block(io.BytesIO(f.read(self.index_block_size)), doc_size=0,
                                min_block_size=self.index_block_size)

    def add_file(self, fd, name, inflate=False):
        """
        Добавляет файл в контейнер

        :param fd: file-like объект файла
        :type fd: BufferedReader
        :param name: Имя файла в контейнере
        :type name: string
        :param inflate: флаг сжатия
        :type inflate: bool
        """
        doc = Document(self)
        attribute_doc_offset = doc.write_header(fd, name)

        doc = Document(self)
        if inflate:
            data_doc_offset = doc.compress(fd, self.file)
        else:
            data_doc_offset = doc.write(fd, min_block_size=self.default_block_size)

        self.toc.append((attribute_doc_offset, data_doc_offset))

    @staticmethod
    def int2hex(value):
        """
        Получает строковое представление целого числа в шестнадцатиричном формате длиной не менее 4 байт

        :param value: конвертируемое число
        :type value: int
        :param size: размер строкового значения
        :type value: int
        :return: предоставление числа
        :type: string
        """
        return f'{value:02x}'.rjust(8, '0')


class Container64(Container):
    end_marker = 0xffffffffffffffff  # 18446744073709551615
    header_fmt = '1Q3i'
    header_size = 20
    block_header_fmt = '2s16s1s16s1s16s1s2s'
    block_header_fmt_size = 16
    block_header_size = 55
    index_fmt = 'Q'
    offset_const = 0x1359
    index_block_size = 0x10000

    @staticmethod
    def int2hex(value):
        """
        Получает строковое представление целого числа в шестнадцатиричном формате длиной не менее 4 байт

        :param value: конвертируемое число
        :type value: int
        :param size: размер строкового значения
        :type value: int
        :return: предоставление числа
        :type: string
        """
        return f'{value:02x}'.rjust(16, '0').upper()



from ..MetaDataObject.core.Container import Container


class Report(Container):
    ext_code = {
        'mgr': '2',  # модуль менеджера Отчета
        'obj': '0',  # модуль Отчета
    }
    help_file_number = 1

    @classmethod
    def get_decode_header(cls, header_data):
        return header_data[0][1][3][1]



from enum import Enum

from .... import helper
from ....ext_exception import ExtException


class FormItemTypes(Enum):
    Field = '381ed624-9217-4e63-85db-c4c3cb87daae'
    CheckBox = '35af3d93-d7c7-4a2e-a8eb-bac87a1a3f26'
    RadioBtn = '782e569a-79a7-4a4f-a936-b48d013936ec'
    SelectField = '64483e7f-3833-48e2-8c75-2c31aac49f6e'
    CommandPanel = 'e69bf21d-97b2-4f37-86db-675aea9ec2cb'
    Button = '6ff79819-710e-4145-97cd-1618da79e3e2'
    Image = '151ef23e-6bb2-4681-83d0-35bc2217230c'
    Group = '90db814a-c75f-4b54-bc96-df62e554d67d'
    Table = 'ea83fe3a-ac3c-4cce-8045-3dddf35b28b1'
    TableField = '236a17b3-7f44-46d9-a907-75f9cdc61ab5'
    Panel = '09ccdc77-ea1a-4a6d-ab1c-3435eada2433'
    Label = '0fc7e20d-f241-460c-bdf4-5ad88e5474a5'
    ListField = '19f8b798-314e-4b4e-8121-905b2a7a03f5'
    Separator = '36e52348-5d60-4770-8e89-a16ed50a2006'
    FieldHtml = 'd92a805c-98ae-4750-9158-d9ce7cec2f20'
    Indicator = 'b1db1f86-abbb-4cf0-8852-fe6ae21650c2'
    CalendarBox = 'e3c063d8-ef92-41be-9c89-b70290b5368b'
    TrackBar = '6c06cd5d-8481-4b6f-a90a-7a97a8bb8bef'
    TextDocumentField = '14c4a229-bfc3-42fe-9ce1-2da049fd0109'
    GraphicalSchemaField = '42248403-7748-49da-b782-e4438fd7bff3'
    GeographicalSchemaField = 'ad37194e-555e-4305-b718-5dca84baf145'
    Chart = 'a8b97779-1a4b-4059-b09c-807f86d2a461'
    GanttChart = 'e5fdc112-5c84-4a16-9728-72b85692b6e2'
    PivotChart = 'a26da99e-184a-4823-b0d6-62816d38dc4e'
    Dendrogram = '984981b1-622d-4ebc-94f7-885f0cdfb59a'


class FormElement:
    name = 'elements'

    @classmethod
    def decode_list(cls, form, raw_data, index_element_count=0, path=''):
        result = []
        element_count = int(raw_data[index_element_count])
        if not element_count:
            return

        for i in range(element_count):
            elem_raw_data = raw_data[index_element_count + i + 1]
            result.append(cls.decode(form, path, elem_raw_data))

        raw_data[index_element_count] = 'Дочерние элементы отдельно'
        del raw_data[index_element_count + 1:index_element_count + 1 + element_count]
        return result

    @classmethod
    def decode(cls, form, path, elem_raw_data):
        metadata_type_uuid = elem_raw_data[0]
        name = helper.str_decode(elem_raw_data[-2][1])
        try:
            metadata_type = FormItemTypes(metadata_type_uuid)
        except ValueError:
            raise ExtException(
                message='Неизвестный тип элемента формы',
                detail=f'{metadata_type_uuid} {name} - {form.__class__.__name__} {form.header["name"]}',
                action='Form802Element.decode'
            )
        form.elements_data[f'{path}/{name}'] = elem_raw_data
        elem_data = dict(
            name=name,
            type=metadata_type.name,
            # raw=raw_data
        )
        return elem_data

    @classmethod
    def encode_list(cls, form, src_dir, file_name, version, raw_data, path=''):
        result = []
        index_element_count = 0
        if raw_data[index_element_count] == 'Дочерние элементы отдельно':
            items = helper.json_read(src_dir, f'{file_name}.elements.tree{version}.json')
            form.elements_data = helper.json_read(src_dir, f'{file_name}.elements.data{version}.json')
            if items:
                for item in items:
                    result.append(cls.encode(form, path, item))
                raw_data[index_element_count] = str(len(items))
                raw_data[index_element_count + 1:index_element_count + 1] = result
            else:
                raw_data[index_element_count] = '0'

    @classmethod
    def encode(cls, form, path, data):
        return form.elements_data[f"{path}/{data['name']}"]


class FormProps(FormElement):
    name = 'props'

    @classmethod
    def decode(cls, form, path, elem_raw_data):
        elem_data = dict(
            name=helper.str_decode(elem_raw_data[4]),
            raw=elem_raw_data
        )
        return elem_data

    @classmethod
    def encode(cls, form, path, data):
        return data['raw']

    @classmethod
    def encode_list(cls, form, src_dir, file_name, version, raw_data, path=''):
        result = []
        index_element_count = 0
        if raw_data[index_element_count] == 'Дочерние элементы отдельно':
            items = helper.json_read(src_dir, f'{file_name}.{cls.name}{version}.json')
            if items:
                for item in items:
                    result.append(cls.encode(form, path, item))
                raw_data[index_element_count] = str(len(items))
                raw_data[index_element_count + 1:index_element_count + 1] = result
            else:
                raw_data[index_element_count] = '0'


from ..MetaDataObject.core.Simple import Simple


class CommonCommand(Simple):
    ext_code = {'obj': 2}
    help_file_number = 1

    @classmethod
    def get_decode_header(cls, header_data):
        return header_data[0][1][1][2][9]



import os
import sys
import unittest

sys.path.append("../../src/")
from v8unpack import helper
from v8unpack.json_container_decoder import JsonContainerDecoder
from v8unpack.unittest_helper import compare_file, NotEqualLine


class TestDecodeBraceFile(unittest.TestCase):
    def setUp(self) -> None:
        self.current_dir = os.path.dirname(__file__)
        self.data_dir = os.path.join(self.current_dir, 'data/json_decode_src')
        self.temp_dir = os.path.join(self.data_dir, 'temp')
        self.temp_dir2 = os.path.join(self.data_dir, 'temp2')

    def test_decode_all_raw_data(self):
        test_files = os.listdir(self.data_dir)
        result = ''
        for file_name in test_files:
            if file_name == 'temp':
                continue
            uuid, x = file_name.split('.')
            result += self.decode_raw_data(uuid)
        self.assertEqual('', result)

    def test_decode_file(self):
        uuid = '66f22eef-9167-4661-acc1-769de41ab428.0'
        result = self.decode_raw_data(uuid)
        self.assertEqual('', result)

    def decode_raw_data(self, file_name):
        helper.clear_dir(self.temp_dir)
        data = helper.brace_file_read(self.data_dir, file_name)
        helper.brace_file_write(data, self.temp_dir, file_name)
        problems = ''
        try:
            result = compare_file(
                os.path.join(self.data_dir, file_name),
                os.path.join(self.temp_dir, file_name),
                problems
            )
        except NotEqualLine as err:
            result = err
        return result



import os
import shutil
from datetime import datetime

from . import helper
from .MetaObject.Configuration802 import Configuration802
from .MetaObject.Configuration803 import Configuration803
from .MetaObject.ConfigurationExtension803 import ConfigurationExtension803
from .MetaObject.ExternalDataProcessor801 import ExternalDataProcessor801
from .MetaObject.ExternalDataProcessor802 import ExternalDataProcessor802
from .MetaObject.ExternalDataProcessor803 import ExternalDataProcessor803
from .ext_exception import ExtException
from .metadata_types import MetaDataTypes

available_types = {
    'ExternalDataProcessor801': ExternalDataProcessor801,
    'ExternalDataProcessor802': ExternalDataProcessor802,
    'ExternalDataProcessor803': ExternalDataProcessor803,
    'Configuration802': Configuration802,
    'Configuration803': Configuration803
}


class Decoder:
    @staticmethod
    def detect_version(src_dir, options=None):
        def init_handler(handler):
            options['version'] = handler.version
            return handler(options=options)

        if os.path.isfile(os.path.join(src_dir, 'root')):
            version = helper.brace_file_read(src_dir, 'version')
            if int(version[0][0][0]) >= 216:
                _tmp = len(version[0][0])
                root = helper.brace_file_read(src_dir, 'root')
                header = helper.brace_file_read(src_dir, root[0][1])
                obj_type = MetaDataTypes(header[0][3][0])
                if _tmp == 2:
                    obj_version = '802'
                elif _tmp == 3:
                    obj_version = version[0][0][2][0][:3]
                    if len(obj_version) == 3:
                        pass
                    elif obj_version in ['1', '2']:
                        obj_version = '802'
                    else:
                        raise Exception(f'Not supported version {obj_version}')
                else:
                    raise Exception(f'Not supported version {_tmp}')
                try:
                    return init_handler(available_types[f'{obj_type.name}{obj_version}'])
                except KeyError:
                    raise Exception(f'Not supported version {obj_type.name}{obj_version}')
            elif version[0][0][0] == "106":
                return init_handler(ExternalDataProcessor801)
        if os.path.isfile(os.path.join(src_dir, 'configinfo')):
            version = helper.brace_file_read(src_dir, 'configinfo')
            if version[0][1][0] == "216":
                if len(version[0][1]) == 3:
                    obj_version = version[0][1][2][0]
                    if obj_version[:3] == '803':
                        return init_handler(ConfigurationExtension803)
        raise Exception('Не удалось определить парсер')

    @classmethod
    def decode(cls, src_dir, dest_dir, *, pool=None, options=None):
        begin = datetime.now()
        print(f'{"Разбираем объект":30}')
        decoder = cls.detect_version(src_dir, options=options)
        helper.clear_dir(dest_dir)
        tasks = decoder.decode(src_dir, dest_dir)  # возвращает список вложенных объектов MetaDataObject
        while tasks:  # многопоточно рекурсивно декодируем вложенные объекты MetaDataObject
            tasks = helper.run_in_pool(cls.decode_include, tasks, pool, title=f'{"Разбираем вложенные объекты":30}',
                                       need_result=True)
            # tasks = helper.list_merge(*tasks_list)
        print(f'{"Разбор объекта закончен":30}: {datetime.now() - begin}')

    @classmethod
    def decode_include(cls, params):
        include_type, (obj_uuid, src_dir, dest_dir, new_dest_path, options) = params
        try:
            handler = helper.get_class_metadata_object(include_type)
            # handler = handler.get_version(decode_params[4])
            tasks = handler.decode(obj_uuid, src_dir, dest_dir, new_dest_path, options, parent_type=include_type)
            return tasks
        except ExtException as err:
            raise ExtException(
                parent=err,
                action=f'{cls.__name__}.decode_include {include_type}'
            )
        except Exception as err:
            raise ExtException(parent=err, action=f'{cls.__name__}.decode_include')

    @classmethod
    def encode(cls, src_dir, dest_dir, *, pool=None, file_name=None, options=None):
        begin = datetime.now()
        print(f'{"Собираем объект":30}')
        helper.clear_dir(dest_dir)
        encoder = cls.get_encoder(src_dir, options)
        # возвращает список вложенных объектов MetaDataObject
        helper.clear_dir(dest_dir)

        parent_id = encoder.get_class_name_without_version()
        child_tasks = encoder.encode_includes(src_dir, file_name, dest_dir, parent_id)
        include_index = {}
        file_list = []
        object_task = []
        title = f'{"Собираем вложенные объекты":30}'
        while object_task or child_tasks:  # многопоточно рекурсивно декодируем вложенные объекты MetaDataObject
            _file_list, _include_index, _object_task, child_tasks = helper.run_in_pool_encode_include(
                cls.encode_include, child_tasks, pool,
                title=title
            )
            if _file_list:
                file_list.extend(_file_list)
            if _include_index:
                helper.update_dict(include_index, _include_index)
            if _object_task:
                object_task.append(_object_task)
            if not child_tasks and object_task:
                _object_task = object_task.pop()
                for elem in _object_task:
                    elem[1][5] = include_index.pop(f"{elem[1][4]}/{elem[0]}/{elem[1][1]}")
                child_tasks = _object_task
                title = f'{"Собираем вложенные объекты":30}'
            a = 1

        encoder.encode(src_dir, dest_dir, file_name=file_name, include_index=include_index.pop(parent_id, None),
                       file_list=file_list)

        print(f'{"Сборка объекта закончена":30}: {datetime.now() - begin}')

    @classmethod
    def get_encoder(cls, src_dir, options=None):
        version = helper.get_options_param(options, 'version', '803')[:3]
        if os.path.isfile(os.path.join(src_dir, 'ExternalDataProcessor.json')):
            versions = {
                '801': ExternalDataProcessor801,
                '802': ExternalDataProcessor802,
                '803': ExternalDataProcessor803,
            }
            return versions[version](options=options)
        if version == '803':
            if os.path.isfile(os.path.join(src_dir, 'ConfigurationExtension.json')):
                return ConfigurationExtension803(options=options)
            if os.path.isfile(os.path.join(src_dir, 'Configuration.json')):
                return Configuration803(options=options)
        elif version == '802':
            if os.path.isfile(os.path.join(src_dir, 'Configuration802.json')):
                return Configuration802(options=options)
        raise FileNotFoundError(f'Не найдены файлы для сборки версии {version} ({src_dir})')

    @classmethod
    def encode_include(cls, params):
        include_type, (new_src_dir, entry, dest_dir, options, parent_id, include_index) = params
        try:
            handler = helper.get_class_metadata_object(include_type)
            handler = handler.get_version(options.get('version'))(options=options)
            handler.title = include_type
            object_task, child_tasks = handler.encode(new_src_dir, entry, dest_dir, parent_id, include_index)
            return object_task, child_tasks
        except Exception as err:
            raise ExtException(parent=err, action=f'Decoder.encode_include({include_type})') from err


def encode(src_dir, dest_dir, *, pool=None, options=None, file_name=None):
    def unpack_dummy():
        helper.clear_dir(dest_dir)
        dummy_path = os.path.join(src_dir, 'dummy.zip')
        if version and int(version.ljust(5, '0')) >= 80316 and os.path.isfile(dummy_path):
            helper.clear_dir(container_dest_dir)
            shutil.unpack_archive(dummy_path, container_dest_dir)
            return os.path.join(dest_dir, '1')
        return container_dest_dir

    version = helper.get_options_param(options, 'version', '803')

    container_dest_dir = os.path.join(dest_dir, '0')
    container_dest_dir = unpack_dummy()
    options = helper.set_options_param(options, 'version', version[:3])
    Decoder.encode(src_dir, container_dest_dir, pool=pool, options=options, file_name=file_name)


def decode(src_dir, dest_dir, *, pool=None, options=None):
    containers = os.listdir(src_dir)
    containers_count = len(containers)
    if containers_count not in [1, 2]:
        raise NotImplementedError(f'Количество контейнеров {containers_count}')

    _src_dir = os.path.join(src_dir, containers[-1])
    Decoder.decode(_src_dir, dest_dir, pool=pool, options=options)
    if containers_count == 2:
        shutil.make_archive(os.path.join(dest_dir, 'dummy'), 'zip', os.path.join(src_dir, '0'))



import os
import shutil
from base64 import b64decode
from enum import Enum

from ..core.Simple import Simple
from ... import helper
from ...ext_exception import ExtException, ExtNotImplemented
from ...json_container_decoder import BigBase64


class TmplType(Enum):
    table = "0"
    base64 = "1"
    active_doc = "2"
    html = "3"
    text = "4"
    geographic = "5"
    scheme = "6"
    design = "7"  # Макет оформления компоновки данных
    graphic_scheme = "8"
    extension = "9"
    # todo добавить остальные типы макетов и их сериализацию


class Template8x(Simple):

    def __init__(self, *, obj_name=None, options=None):
        super().__init__(obj_name=obj_name, options=options)
        self.data = None
        self.tmpl_type = None
        self.raw_data = None
        self.raw_header = None

    @property
    def uuid(self):
        if self.header:
            return self.header['uuid']
        return None

    def decode_object(self, src_dir, uuid, dest_dir, dest_path, version, header):
        try:
            super(Template8x, self).decode_object(src_dir, uuid, dest_dir, dest_path, version, header)
            tmpl_type = self.get_template_type(header)
            try:
                self.tmpl_type = TmplType(tmpl_type)
            except Exception as err:
                raise ExtNotImplemented(message='Неизвестный тип макета', detail=f'{tmpl_type} у {self.header.name} в файле {uuid}')
            self.header['type'] = self.tmpl_type.name

            _dest_dir = os.path.join(dest_dir, dest_path)
        except Exception as err:
            raise ExtException(parent=err)
        try:
            getattr(self, f'decode_{self.tmpl_type.name}_data')(src_dir, _dest_dir, True)
        except AttributeError:
            raise Exception(f'Не реализованный тип макета {self.header["type"]}')

    @classmethod
    def get_decode_header(cls, header_data):
        return header_data[0][1][2]

    @classmethod
    def get_template_type(cls, header_data):
        return header_data[0][1][1]

    def decode_code(self, src_dir):
        return

    def decode_includes(self, src_dir, dest_dir, dest_path, header):
        return []

    def decode_active_doc_data(self, src_dir, dest_dir, write):
        self.decode_scheme_data(src_dir, dest_dir, write)

    def decode_geographic_data(self, src_dir, dest_dir, write):
        self.decode_scheme_data(src_dir, dest_dir, write)

    def decode_design_data(self, src_dir, dest_dir, write):
        self.decode_scheme_data(src_dir, dest_dir, write)

    def decode_graphic_scheme_data(self, src_dir, dest_dir, write):
        self.decode_scheme_data(src_dir, dest_dir, write)

    def decode_scheme_data(self, src_dir, dest_dir, write, *, extension='bin'):
        try:
            shutil.copy2(
                os.path.join(src_dir, f'{self.header["uuid"]}.0'),
                os.path.join(dest_dir, f'{self.header["name"]}.{extension}')
            )
            # self.data = helper.bin_read(src_dir, f'{self.header["uuid"]}.0')
            # if write:
            #     helper.bin_write(self.data, dest_dir, f'{self.header["name"]}.bin')
        except FileNotFoundError:
            return
            # self.decode_text_data(src_dir, dest_dir, write) ??? что это

    def decode_table_data(self, src_dir, dest_dir, write):
        self.decode_scheme_data(src_dir, dest_dir, write, extension='mxl')


    def decode_text_data(self, src_dir, dest_dir, write):
        try:
            self.data, encoding = helper.txt_read_detect_encoding(src_dir, f'{self.header["uuid"]}.0')
        except FileNotFoundError:
            return
        if write:
            helper.txt_write(self.data, dest_dir, f'{self.header["name"]}.txt', encoding=encoding)

    def decode_extension_data(self, src_dir, dest_dir, write):
        self.decode_base64_data(src_dir, dest_dir, write)

    def decode_base64_data(self, src_dir, dest_dir, write):
        filename = f'{self.header["uuid"]}.0'
        try:
            data = helper.brace_file_read(src_dir, filename)
        except BigBase64:
            shutil.copy2(os.path.join(src_dir, filename), os.path.join(dest_dir, f'{self.header["name"]}.c1b64'))
            return
        except FileNotFoundError:
            file_name = f'{self.header["uuid"]}.0'
            if os.path.isfile(os.path.join(src_dir, file_name)):
                # todo сюда можно вставить выдергивание бинарника из файла если кому то хочется
                shutil.copy2(
                    os.path.join(src_dir, file_name),
                    os.path.join(dest_dir, f'{self.header["name"]}')
                )
            return
        if data[0][1] and data[0][1][0]:
            self.data = b64decode(data[0][1][0][8:])
            data[0][1][0] = '"данные в отдельном файле"'
            if write:
                extension = helper.get_extension_from_comment(self.header['comment'])
                helper.bin_write(self.data, dest_dir, f'{self.header["name"]}.{extension}')

    def decode_html_data(self, src_dir, dest_dir, write):
        self._decode_html_data(src_dir, dest_dir, self.new_dest_file_name)

    def decode_header(self, header):
        self.tmpl_type = TmplType(header[0][1][1])
        self.header = {
            'type': self.tmpl_type.name,
        }
        _header = header[0][1][2]
        helper.decode_header(self.header, _header)

    def encode_object(self, src_dir, file_name, dest_dir):
        self.tmpl_type = TmplType[self.header['type']]

        self.raw_header = self.encode_header()
        try:
            handler = f'encode_{self.tmpl_type.name}_data'
            getattr(self, handler)(src_dir, dest_dir)
        except AttributeError:
            raise Exception(f'Не реализованный тип макета {self.header["type"]}')

        # helper.json_write(self.raw_header, dest_dir, f'{self.header["uuid"]}.bin')

    def encode_active_doc_data(self, src_dir, dest_dir):
        self.encode_scheme_data(src_dir, dest_dir)

    def encode_table_data(self, src_dir, dest_dir):
        self.encode_scheme_data(src_dir, dest_dir, extension='mxl')

    def encode_geographic_data(self, src_dir, dest_dir):
        self.encode_scheme_data(src_dir, dest_dir)

    def encode_design_data(self, src_dir, dest_dir):
        self.encode_scheme_data(src_dir, dest_dir)

    def encode_graphic_scheme_data(self, src_dir, dest_dir):
        self.encode_scheme_data(src_dir, dest_dir)

    def encode_scheme_data(self, src_dir, dest_dir, *, extension='bin'):
        try:
            file_name = f'{self.header["uuid"]}.0'
            shutil.copy2(
                os.path.join(src_dir, f'{self.header["name"]}.{extension}'),
                os.path.join(dest_dir, file_name)
            )
            self.file_list.append(file_name)
        except FileNotFoundError:
            self.encode_text_data(src_dir, dest_dir)

    def encode_text_data(self, src_dir, dest_dir):
        try:
            raw_data, encoding = helper.txt_read_detect_encoding(src_dir, f'{self.header["name"]}.txt')
            file_name = f'{self.header["uuid"]}.0'
            helper.txt_write(raw_data, dest_dir, file_name, encoding=encoding)
            self.file_list.append(file_name)
        except FileNotFoundError:
            pass

    def encode_extension_data(self, src_dir, dest_dir):
        self.encode_base64_data(src_dir, dest_dir)

    def encode_base64_data(self, src_dir, dest_dir):
        extension = helper.get_extension_from_comment(self.header['comment'])
        try:
            bin_data = helper.bin_read(src_dir, f'{self.header["name"]}.{extension}')
            self._encode_bin_data(bin_data, dest_dir)
        except FileNotFoundError:
            file_name = f'{self.header["name"]}.c1b64'
            if os.path.isfile(os.path.join(src_dir, file_name)):
                dest_file_name = f'{self.header["uuid"]}.0'
                self.file_list.append(dest_file_name)
                shutil.copy2(
                    os.path.join(src_dir, file_name),
                    os.path.join(dest_dir, dest_file_name)
                )

    def encode_html_data(self, src_dir, dest_dir):
        self._encode_html_data(src_dir, self.header["name"], dest_dir)

    def _encode_bin_data(self, bin_data, dest_dir):
        self.raw_data = [[
            "1",
            [self._get_b64_string(bin_data)]
        ]]

        file_name = f'{self.header["uuid"]}.0'
        helper.brace_file_write(self.raw_data, dest_dir, file_name)
        self.file_list.append(file_name)

    def encode_header(self):
        raise NotImplemented()



from ..version import __version__
from .. import helper
from ..MetaObject import MetaObject
from ..ext_exception import ExtException


class ExternalDataProcessor(MetaObject):

    def __init__(self, *, obj_name=None, options=None):
        super().__init__(obj_name=obj_name, options=options)
        self.data = None

    def decode(self, src_dir, dest_dir):
        self.header = {}
        self.data = {}
        root = helper.brace_file_read(src_dir, 'root')
        self.header["file_uuid"] = root[0][1]
        _header_data = helper.brace_file_read(src_dir, f'{self.header["file_uuid"]}')
        self.set_header_data(_header_data)

        root = helper.brace_file_read(src_dir, 'root')
        self.header['v8unpack'] = __version__
        self.header['file_uuid'] = root[0][1]
        self.data['version'] = helper.brace_file_read(src_dir, 'version')
        # self.header['versions'] = helper.brace_file_read(src_dir, 'versions')
        self.data['copyinfo'] = helper.brace_file_read(src_dir, 'copyinfo')

        try:
            form1 = helper.brace_file_read(src_dir, f'{self.header["uuid"]}.1')
        except FileNotFoundError:
            form1 = None

        self.data['form1'] = form1

        self.decode_code(src_dir)
        pass
        _file_name = self.get_class_name_without_version()

        tasks = self.decode_includes(src_dir, dest_dir, '', self.header['data'])

        helper.json_write(self.header, dest_dir, f'{_file_name}.json')
        helper.json_write(self.data, dest_dir, f'{_file_name}.data{self.version}.json')
        self.write_decode_code(dest_dir, 'ExternalDataProcessor')

        return tasks
        # helper.run_in_pool(self.decode_include, tasks, pool)
        pass

    @classmethod
    def get_decode_includes(cls, header_data):
        return [header_data[0][3][1]]

    @classmethod
    def get_decode_header(cls, header_data):
        return header_data[0][3][1][1][3][1]

    def encode(self, src_dir, dest_dir, *, file_name=None, include_index=None, file_list=None, **kwargs):
        try:
            _file_name = self.get_class_name_without_version()
            self.header = helper.json_read(src_dir, f'{_file_name}.json')
            helper.check_version(__version__, self.header.get('v8unpack', ''))
            try:
                self.data = helper.json_read(src_dir, f'{_file_name}.data{self.version}.json')
            except FileNotFoundError:
                self.data = self.encode_empty_data()

            self.set_product_info(src_dir, file_name)

            if include_index and self.get_options('auto_include'):
                self.fill_header_includes(include_index)

            helper.brace_file_write(self.encode_root(), dest_dir, 'root')
            file_list.append('root')
            helper.brace_file_write(self.encode_version(), dest_dir, 'version')
            file_list.append('version')
            helper.brace_file_write(self.data['copyinfo'], dest_dir, 'copyinfo')
            file_list.append('copyinfo')
            # helper.brace_file_write(self.header['versions'], dest_dir, 'versions')
            helper.brace_file_write(self.header['data'], dest_dir, self.header["file_uuid"])
            file_list.append(self.header["file_uuid"])
            if self.data.get('form1'):
                file_name = f'{self.header["uuid"]}.1'
                helper.brace_file_write(self.data['form1'], dest_dir, file_name)
                file_list.append(file_name)
            self.encode_code(src_dir, 'ExternalDataProcessor')
            self.write_encode_code(dest_dir)

            file_list.append('versions')
            file_list.extend(self.file_list)
            versions = self.encode_versions(file_list)
            helper.brace_file_write(versions, dest_dir, 'versions')
            return None
        except Exception as err:
            raise ExtException(parent=err)

    def encode_root(self):
        return [[
            "2",
            self.header["file_uuid"],
            ""
        ]]

    def encode_empty_data(self):
        return {
            "copyinfo": [
                [
                    "4",
                    [
                        "0"
                    ],
                    [
                        "0"
                    ],
                    [
                        "0"
                    ],
                    [
                        "0",
                        "0"
                    ],
                    [
                        "0"
                    ]
                ]
            ]
        }



from ..MetaDataObject.core.Container import Container


class ChartOfCalculationTypes(Container):
    ext_code = {
        'obj': 0,
        'mgr': 3
    }
    help_file_number = 1
    predefined_file_number = 2

    @classmethod
    def get_decode_header(cls, header):
        return header[0][1][1][1]



from ..MetaDataObject.Form import Form
from ..MetaDataObject.versions.DocumentForm803 import DocumentForm803 as Form803
from ..MetaDataObject.versions.Form801 import Form801
from ..MetaDataObject.versions.Form802 import Form802
from ..MetaDataObject.versions.OldForm803 import OldForm801, OldForm802


class DocumentForm(Form):
    versions = {
        '801': Form801,
        '802': Form802,
        '803': Form803,
        '5-5': OldForm801,
        '7-7': OldForm801,
        '9-9': OldForm802,
        '0-5': Form801,
        '0-7': Form801,
        '0-9': Form802,
        '0-12': Form803,
        '0-13': Form803,
        '1-9': Form802,
        '1-12': Form803,
        '1-13': Form803
    }



import argparse
import os
import shutil
import sys
import tempfile
from datetime import datetime

from . import helper
from .container_reader import extract as container_extract, decompress_and_extract
from .container_writer import build as container_build, compress_and_build
from .decoder import decode, encode
from .ext_exception import ExtException
from .file_organizer import FileOrganizer
from .file_organizer_ce import FileOrganizerCE
from .helper import check_index, load_json
from .index import update_index
from .version import __version__


def extract(in_filename: str, out_dir_name: str, *, temp_dir=None, index=None, processes=None, options=None):
    try:
        begin0 = datetime.now()
        print(f"v8unpack {__version__}")

        index = check_index(index)
        if not index and index is not None:
            return

        print(f"Начали")

        descent = options.get('descent') if options else None

        if descent is None:
            helper.clear_dir(os.path.normpath(out_dir_name))
        clear_temp_dir = False
        if temp_dir is None:
            clear_temp_dir = True
            temp_dir = tempfile.mkdtemp()
        helper.clear_dir(os.path.normpath(temp_dir))

        dir_stage0 = os.path.join(temp_dir, 'decode_stage_0')
        dir_stage1 = os.path.join(temp_dir, 'decode_stage_1')
        # dir_stage2 = os.path.join(temp_dir, 'decode_stage_2')
        dir_stage3 = os.path.join(temp_dir, 'decode_stage_3')

        pool = helper.get_pool(processes=processes)

        container_extract(in_filename, dir_stage0, False, False)
        decompress_and_extract(dir_stage0, dir_stage1, pool=pool)

        # json_decode(dir_stage1, dir_stage2, pool=pool)

        decode(dir_stage1, dir_stage3, pool=pool, options=options)

        if descent is not None:
            FileOrganizerCE.unpack(dir_stage3, out_dir_name, pool=pool, index=index, descent=descent)
        else:
            FileOrganizer.unpack(dir_stage3, out_dir_name, pool=pool, index=index)

        end = datetime.now()
        print(f'{"Готово":30}: {end - begin0}')

        helper.close_pool(pool)
        if clear_temp_dir:
            shutil.rmtree(temp_dir, ignore_errors=True)
    except Exception as err:
        error = ExtException(parent=err)
        print(f'\n\n{error}')


def build(in_dir_name: str, out_file_name: str, *, temp_dir=None, index=None,
          options=None, processes: int = None):
    try:
        begin0 = datetime.now()

        print(f"v8unpack {__version__}")

        index = check_index(index)
        if not index and index is not None:
            return

        print(f"Начали")

        clear_temp_dir = False
        if temp_dir is None:
            clear_temp_dir = True
            temp_dir = tempfile.mkdtemp()
        helper.clear_dir(os.path.normpath(temp_dir))

        dir_stage0 = os.path.join(temp_dir, 'encode_stage_0')
        dir_stage1 = os.path.join(temp_dir, 'encode_stage_1')
        # dir_stage2 = os.path.join(temp_dir, 'encode_stage_2')
        dir_stage3 = os.path.join(temp_dir, 'encode_stage_3')

        pool = helper.get_pool(processes=processes)

        descent = options.get('descent') if options else None
        if descent is None:
            FileOrganizer.pack(in_dir_name, dir_stage3, pool=pool, index=index)
        else:
            FileOrganizerCE.pack(in_dir_name, dir_stage3, pool=pool, index=index, descent=descent)

        encode(dir_stage3, dir_stage1, pool=pool, file_name=os.path.basename(out_file_name), options=options)

        # json_encode(dir_stage2, dir_stage1, pool=pool)

        compress_and_build(dir_stage1, dir_stage0, pool=pool)
        container_build(dir_stage0, out_file_name, True)

        helper.close_pool(pool)
        if clear_temp_dir:
            shutil.rmtree(temp_dir, ignore_errors=True)

        end = datetime.now()
        print(f'{"Готово":30}: {end - begin0}')

    except Exception as err:
        error = ExtException(parent=err)
        print(f'\n\n{error}')


def build_all(product_file_name: str, product_code: str = None, processes=None):
    try:
        products = load_json(os.path.abspath(product_file_name))
        if product_code:
            products = {product_code: products[product_code]}
            products[product_code]['disable'] = False
        for product, params in products.items():
            if params.get('disable'):
                continue
            print(f'\nСобираем {product}')
            build(
                params['src'], params['bin'],
                temp_dir=params.get('temp'), index=params.get('index'),
                options=params.get('options'), processes=processes
            )
        pass
    except Exception as err:
        error = ExtException(parent=err)
        print(f'\n\n{error}')


def extract_all(product_file_name: str, product_code: str = None, processes=None):
    try:
        products = load_json(os.path.abspath(product_file_name))
        if product_code:
            products = {product_code: products[product_code]}
            products[product_code]['disable'] = False
        for product, params in products.items():
            if params.get('disable'):
                continue
            print(f'\nРазбираем {product}\n')
            extract(
                params['bin'], params['src'],
                temp_dir=params.get('temp'), index=params.get('index'),
                options=params.get('options'), processes=processes
            )
        pass
    except Exception as err:
        error = ExtException(parent=err)
        print(f'\n\n{error}')


def update_index_all(product_file_name: str, product_code: str = None, dest_dir: str = None):
    try:
        products = load_json(os.path.abspath(product_file_name))
        if product_code:
            products = {product_code: products[product_code]}
            products[product_code]['disable'] = False
        for product, params in products.items():
            if params.get('disable'):
                continue
            print(f'\nФормируем индекс {product}\n')
            update_index(
                params['src'],
                params.get('index'),
                dest_dir
            )
        pass
    except Exception as err:
        error = ExtException(parent=err)
        print(f'\n\n{error}')


def main():
    parser = argparse.ArgumentParser(
        prog=f'v8unpack {__version__}',
        description='Распаковка и сборка бинарных файлов 1С'
    )
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-E', nargs=2, metavar=('file', 'src'),
                       help='разобрать файл 1С, где '
                            'file - путь до бинарного файла, '
                            'src -  папка куда будут помещены исходники')
    group.add_argument('-B', nargs=2, metavar=('src', 'file'),
                       help='собрать файл 1С, где '
                            'src - путь до папки с исходниками, '
                            'file - путь до бинарного файла')
    group.add_argument('-I', nargs=1, metavar='src',
                       help='сформировать index, где '
                            'src - путь до папки с исходниками'
                       )
    parser.add_argument('--temp', help='путь до временной папки')
    parser.add_argument('--core', help='название общей папки добавляемой в индекс по умолчанию, только для режима -I')
    parser.add_argument('--index', help='путь до json файла с словарем копирования,'
                                        'структура файла: {путь исходника: путь общей папки}')
    parser.add_argument('--version', default='803',
                        help="версия сборки, для сборки обработок указывается версия платформы 801/802/803, "
                             " для сборки расширений указывается версия режима совместимости, "
                             "например для 8.3.6 это 80306, подробности в документации на github")
    parser.add_argument('--processes', type=int,
                        help="количество процессов которые будут использоваться для работы, "
                             " если не указано, то количество процессов = количеству ядер компьютера. "
                             " Нет смысла указывать значения превышающее количество ядер")
    parser.add_argument('--descent', type=int,
                        help="включает режим наследования при сборке и разборке,"
                             "четырех значный формат 3.0.75.100 (не более 3 знаков на каждый разряд)"
                             "подробности в инструкции - раздел разработка расширений")
    parser.add_argument('--gui',
                        help="режим совместимости интерфейса 1С, переопределяет значение из исходников "
                             "для расширений, если указан устанавливается в соответствующий реквизит. "
                             "Допустимые значения: "
                             " 0 - Версия 8.2, 1 - Версия 8.2. Разрешить Такси,"
                             " 2- Такси. Разрешить Версия 8.2, 3 - Такси")
    parser.add_argument("--auto_include", default=False, action="store_true",
                        help="Если указан содержимое метаданных собирается динамически в зависимости "
                             "от наличия файлов в каталоге, а при разборке оглавление не сохраняется, "
                             "например вы можете подкинуть файлы метаданных в каталог из другого проекта "
                             "минусом является, что вложенные ресурсы всегда сортируются по алфавиту. "
                             "Если не указан, то сборщик не трогает оглавление, вы добавляете и убираете"
                             "вложенные объекты всегда через конфигуратор")

    group.add_argument('-EA', nargs=1, metavar='file',
                       help='разобрать один или несколько файлов 1С, где '
                            'file - путь до json файла со списком продуктов и параметрами их сборки')
    group.add_argument('-BA', nargs=1, metavar='file',
                       help='собрать один или несколько файлов 1С, где '
                            'file - путь до json файла со списком продуктов и параметрами их сборки')

    group.add_argument('-IA', nargs=1, metavar='file',
                       help='сформировать index по одному или нескольким продуктам, где '
                            'file - путь до json файла со списком продуктов и параметрами их сборки')

    if len(sys.argv) == 1:
        parser.print_help()
        return 1

    args = parser.parse_args()

    options = {}
    if args.gui:
        options['gui'] = args.gui

    if args.version:
        options['version'] = args.version

    if args.auto_include:
        options['auto_include'] = args.auto_include

    if args.descent:
        options['descent'] = args.descent

    if args.E is not None:
        extract(os.path.abspath(args.E[0]), os.path.abspath(args.E[1]),
                index=args.index, temp_dir=args.temp, options=options)
        return

    if args.B is not None:
        build(os.path.abspath(args.B[0]), os.path.abspath(args.B[1]),
              index=args.index, temp_dir=args.temp, options=options)
        return

    if args.BA is not None:
        build_all(args.BA[0], args.index)
        return

    if args.EA is not None:
        extract_all(args.EA[0], args.index)
        return

    if args.I is not None:
        update_index(args.I[0], args.index, args.core)
        return

    if args.IA is not None:
        update_index_all(args.IA[0], args.index, args.core)
        return


if __name__ == '__main__':
    sys.exit(main())



from ..MetaDataObject.core.Container import Container


class AccountingRegister(Container):
    ext_code = {
        'obj': 6,
        'mgr': 7
    }
    help_file_number = 5

    @classmethod
    def get_decode_header(cls, header):
        return header[0][1][15][1]



import os
import sys
import unittest

sys.path.append("../../src/")
from v8unpack.MetaDataObject.Role import Role as Meta
from v8unpack import helper


class TestFileOrganizerCE(unittest.TestCase):
    def setUp(self) -> None:
        self.current_dir = os.path.dirname(__file__)
        self.data_dir = os.path.join(self.current_dir, 'data', 'json_decode_src')
        self.temp_dir = os.path.join(self.data_dir, 'temp')

    def test_near(self):
        form = Meta()
        file_name = "8a310493-c06f-46db-b2b5-14d6ed203312"
        # os.remove(os.path.join(self.temp_dir, file_name))
        helper.clear_dir(os.path.join(self.temp_dir, file_name))
        form.decode(self.data_dir, file_name, self.temp_dir, file_name, 803)



import os

from .. import helper
from ..MetaObject import MetaObject
from ..ext_exception import ExtException


class MetaDataObject(MetaObject):
    versions = None
    version = None
    help_file_number = None

    def __init__(self, *, obj_name=None, options=None):
        super().__init__(obj_name=obj_name, options=options)
        self.path = ''
        self.new_dest_path = None
        self.new_dest_dir = None
        self.new_dest_file_name = None
        self.obj_version = None

    @classmethod
    def get_obj_name(cls):
        return cls._obj_name if cls._obj_name else cls.__name__

    @classmethod
    def get_version(cls, version):
        if cls.versions is None:
            return cls
        try:
            return cls.versions[version]
        except KeyError:
            raise Exception(f'Нет реализации {cls.__name__} для версии "{version}"')

    @staticmethod
    def brace_file_read(src_dir, file_name):
        return helper.brace_file_read(src_dir, file_name)

    @classmethod
    def decode_get_handler(cls, src_dir, file_name, options):
        try:
            version = helper.get_options_param(options, 'version', '803')
            return cls.get_version(version)(options=options)
        except Exception as err:
            raise ExtException(message='Не смогли получить класс объекта', detail=f'{cls.__name__} {err}')

    @classmethod
    def decode(cls, src_dir: str, file_name: str, dest_dir: str, dest_path: str, options, *, parent_type=None):
        try:
            self = cls.decode_get_handler(src_dir, file_name, options)
            header_data = cls.brace_file_read(src_dir, file_name)
            # self = cls()
            if parent_type:
                self.title = parent_type
            self.decode_object(src_dir, file_name, dest_dir, dest_path, self.version, header_data)
            tasks = self.decode_includes(src_dir, dest_dir, self.new_dest_path, header_data)
            self.write_decode_object(dest_dir, self.new_dest_path, self.new_dest_file_name)
            return tasks
        except Exception as err:
            problem_file = os.path.join(os.path.basename(src_dir), file_name)
            raise ExtException(
                parent=err,
                message="Ошибка декодирования",
                detail=f'объекта метаданных "{cls.__name__}" файл "{problem_file}" ({dest_path})',
                action=f'{cls.__name__}.decode'
            ) from err

    def set_write_decode_mode(self, dest_dir, dest_path):
        self.set_mode_decode_in_root_folder(dest_dir, dest_path)

    def decode_object(self, src_dir, file_name, dest_dir, dest_path, version, header_data):
        self.set_header_data(header_data)
        self.set_write_decode_mode(dest_dir, dest_path)

    def write_decode_object(self, dest_dir, dest_path, file_name):
        dest_full_path = os.path.join(dest_dir, dest_path)
        helper.json_write(self.header, dest_full_path, f'{file_name}.json')
        self.write_decode_code(dest_full_path, file_name)

    def set_mode_decode_in_name_folder(self, dest_dir, dest_path):
        self.new_dest_path = os.path.join(dest_path, self.header['name'])
        self.new_dest_dir = os.path.join(dest_dir, self.new_dest_path)
        self.new_dest_file_name = self.get_obj_name()
        helper.makedirs(self.new_dest_dir)

    def set_mode_decode_in_root_folder(self, dest_dir, dest_path):
        self.new_dest_path = dest_path
        self.new_dest_dir = os.path.join(dest_dir, self.new_dest_path)
        self.new_dest_file_name = self.header['name']

    def get_include_obj_uuid(self):
        return self.header['uuid']

    def encode(self, src_dir, file_name, dest_dir, parent_id, include_index):
        src_file_name = self.get_encode_file_name(file_name)
        try:
            if not include_index:
                current_obj_id = f"{parent_id}/{self.title}/{file_name}"
                child_tasks = self.encode_includes(src_dir, src_file_name, dest_dir, current_obj_id)
                if child_tasks:
                    object_task = [self.__class__.__name__, [src_dir, file_name, dest_dir, self.options, parent_id, {}]]
                    return object_task, child_tasks
            try:
                self.header = helper.json_read(src_dir, f'{src_file_name}.json')
            except FileNotFoundError:
                return
            if include_index and self.get_options('auto_include'):
                self.fill_header_includes(include_index)  # todo dynamic index
            self.encode_object(src_dir, src_file_name, dest_dir)
            self.write_encode_object(dest_dir)
            return dict(
                parent_id=parent_id,
                file_list=self.file_list,
                obj_type=self.title,
                obj_uuid=self.get_include_obj_uuid()
            ), None
        except Exception as err:
            raise ExtException(
                parent=err,
                dump=dict(uuid=self.header['uuid'], src_dir=src_dir, file_name=file_name),
                action=f'{self.__class__.__name__}.encode') from err

    def get_encode_file_name(self, file_name):
        return file_name

    def encode_object(self, src_dir, file_name, dest_dir):
        msg = f'Нет реализации для "{self.__class__.__name__}"'
        raise Exception(msg)

    def write_encode_object(self, dest_dir):
        file_name = f'{self.header["uuid"]}'
        helper.brace_file_write(self.header['data'], dest_dir, file_name)
        self.file_list.append(file_name)
        self.write_encode_code(dest_dir)

    @staticmethod
    def get_metadata_object_version(path, file_name):
        _path = os.path.join(path, file_name)
        with open(_path, 'r', encoding='utf-8-sig') as file:
            header = file.read(9)
            root = header.split('{')
            if len(root) < 3 or root[1][0] != '1':
                raise ExtException(message='Неизвестная версия объекта метаданных',
                                   detail=file_name)
            root = root[2].split(',')
            return int(root[0])



from .Simple import Simple
from .. import helper
import os
import shutil


class SimpleWithInfo(Simple):
    def decode_object(self, src_dir, file_name, dest_dir, dest_path, version, header_data):
        super(Simple, self).decode_object(src_dir, file_name, dest_dir, dest_path, version, header_data)
        try:
            _src = os.path.join(src_dir, f'{self.header["uuid"]}.0')
            _dest = os.path.join(dest_dir, self.new_dest_path, f'{self.new_dest_file_name}.0.c1brace')
            shutil.copy2(_src, _dest)
            # self.header['info'] = helper.brace_file_read(src_dir, f'{self.header["uuid"]}.0')
        except FileNotFoundError:
            return

    # def write_encode_object(self, dest_dir):
    #     super().write_encode_object(dest_dir)
        # info = self.header.get('info')
        # if info:
        #     file_name = f'{self.header["uuid"]}.0'
        #     helper.brace_file_write(info, dest_dir, file_name)
        #     self.file_list.append(file_name)

    def encode_object(self, src_dir, file_name, dest_dir):
        super().encode_object(src_dir, file_name, dest_dir)
        try:
            _src = os.path.join(src_dir, f'{file_name}.0.c1brace')
            _dest_file_name = f'{self.header["uuid"]}.0'
            _dest = os.path.join(dest_dir, _dest_file_name)
            shutil.copy2(_src, _dest)
            self.file_list.append(_dest_file_name)
            # self.header['info'] = helper.brace_file_read(src_dir, f'{self.header["uuid"]}.0')
        except FileNotFoundError:
            return



import json
import os

from .helper import remove_descent_from_filename
from .ext_exception import ExtException


def get(index: dict, path: str, file_name: str):
    try:
        _index = index
        if path:
            _path = path.split('\\')
            for _dir in _path:
                _index = _index[_dir]
        return _index[file_name]
    except TypeError:
        raise ExtException(
            message="Ошибка c уровнями вложенности в Index.json",
            detail=f'{path}\{file_name}',
            action='index.get'
        )


def update_index(src_dir: str, index_file_name: str, dest_dir: str):
    def _update_index(_src_dir, _index, _dest_dir, _path):
        entries = os.listdir(os.path.join(_src_dir, _path))
        for entry in entries:
            entry = remove_descent_from_filename(entry)
            new_path = os.path.join(_path, entry)
            if os.path.isdir(os.path.join(_src_dir, new_path)):
                if entry not in _index:
                    _index[entry] = {}
                _update_index(_src_dir, _index[entry], _dest_dir, new_path)
            else:
                if entry not in _index:
                    _index[entry] = os.path.join(_dest_dir, _path, entry) if _dest_dir else ''
        pass

    try:
        with open(index_file_name, 'r', encoding='utf-8') as f:
            index = json.load(f)
    except FileNotFoundError:
        index = {}

    _update_index(src_dir, index, dest_dir, '')
    with open(index_file_name, 'w+', encoding='utf-8') as f:
        json.dump(index, f, ensure_ascii=False, indent=2)



from ..MetaDataObject.core.Simple import Simple


class CommonAttribute(Simple):
    # ext_code = {
    #     'mgr': '1',  # модуль менеджера Константы
    #     'obj': '0',  # модуль менеджера значения Константы
    # }

    @classmethod
    def get_decode_header(cls, header_data):
        return header_data[0][1][1][1][1]



from .FormElement import FormElement, check_count_element, calc_offset
from ....helper import FuckingBrackets
from ....ext_exception import ExtException


class Table(FormElement):
    @classmethod
    def get_name_node_offset(cls, raw_data):
        return calc_offset([(4, 1), (1, 0)], raw_data)

    @classmethod
    def decode(cls, form, path, raw_data):
        try:
            size = check_count_element([
                (4, 1), (50, 2), (7, 2)
            ], raw_data)
        except Exception as err:
            raise ExtException(parent=err)
        if raw_data[0] == '55' and size != 99:
            raise FuckingBrackets()
        return super().decode(form, path, raw_data)


from ..MetaDataObject import MetaDataObject
import os
from .. import helper


class XDTOPackage(MetaDataObject):

    def decode_object(self, src_dir, file_name, dest_dir, dest_path, version, header_data):
        super().decode_object(src_dir, file_name, dest_dir, dest_path, version, header_data)
        try:
            package = helper.bin_read(src_dir, f'{self.header["uuid"]}.0')
            helper.bin_write(package, self.new_dest_dir, f'{self.new_dest_file_name}.bin')
        except FileNotFoundError:
            return

    def decode_includes(self, src_dir, dest_dir, dest_path, header):
        return []

    def encode_includes(self, src_dir, file_name, dest_dir, parent_id):
        return []

    def encode_object(self, src_dir, file_name, dest_dir):
        try:
            package = helper.bin_read(src_dir, f'{file_name}.bin')
            file_name = f'{self.header["uuid"]}.0'
            helper.bin_write(package, dest_dir, file_name)
            self.file_list.append(file_name)
        except FileNotFoundError:
            return
        return []



from ..MetaDataObject.core.Container import Container


class SettingsStorage(Container):
    ext_code = {
        'mgr': 8,
    }
    # help_file_number = 5

    @classmethod
    def get_decode_header(cls, header):
        return header[0][1][1][1]



from enum import Enum


class MetaDataGroup(Enum):
    General = '9cd510cd-abfc-11d4-9434-004095e12fc7'
    Main = '9fcd25a0-4822-11d4-9414-008048da11f9'
    Accounting = 'e3687481-0a87-462c-a166-9f34594f9bba'
    Calculation = '9de14907-ec23-4a07-96f0-85521cb6b53b'
    DocFlow = '51f2d5d8-ea4d-4064-8892-82951750031e'
    DataSource = 'e68182ea-4237-4383-967f-90c1e3370bc7'
    Integration = 'fb282519-d103-4dd3-bc12-cb271d631dfc'


class MetaDataTypes(Enum):
    # todo в синтаксис помощнике найти все английские названия искать строкой "ОбъектМетаданных: [хxx...]"
    # DocumentAttribute = "45e46cbc-3e24-4165-8b7b-cc98a6f80211"        # Документ.Реквизиты
    # ExternalDataSource = 'e68182ea-4237-4383-967f-90c1e3370bc7'         # Внешние источники данных (не точно...)
    AccountingRegister = '2deed9b8-0056-4ffe-a473-c20a6c32a0bc'  # Регистры бухгалтерии
    AccountingRegisterCommand = "7162da60-f7fe-4d78-ad5d-e31700f9af18"
    AccountingRegisterForm = 'd3b5d6eb-4ea2-4610-a3e2-624d4e815934'  # Регистры бухгалтерии
    AccumulationRegister = 'b64d9a40-1642-11d6-a3c7-0050bae0a776'  # Регистры накопления
    AccumulationRegisterCommand = "99f328af-a77f-4572-a2d8-80ed20c81890"
    AccumulationRegisterForm = 'b64d9a44-1642-11d6-a3c7-0050bae0a776'  # Регистры накопления форма
    Bot = '6e6dc072-b7ac-41e7-8f88-278d25b6da2a'  # Бот
    BusinessProcess = 'fcd3404e-1523-48ce-9bc0-ecdb822684a1'  # Бизнес-процессы
    BusinessProcessCommand = '7a3e533c-f232-40d5-a932-6a311d2480bf'
    BusinessProcessForm = '3f7a8120-b71a-4265-98bf-4d9bc09b7719'
    CalculationRegister = 'f2de87a8-64e5-45eb-a22d-b3aedab050e7'  # Регистры расчета
    CalculationRegisterCommand = "acdf0f11-2d59-4e37-9945-c6721871a8fe"
    CalculationRegisterForm = 'a2cb086c-db98-43e4-a1a9-0760ab048f8d'  # Регистры расчета
    CalculationRegisterRecalculations = '274bf899-db0e-4df6-8ab5-67bf6371ec0b'  # Перерасчеты
    Catalog = "cf4abea6-37b2-11d4-940f-008048da11f9"  # Справочники
    CatalogCommand = '4fe87c89-9ad4-43f6-9fdb-9dc83b3879c6'  # Команда справочника
    CatalogForm = 'fdf816d2-1ead-11d5-b975-0050bae0a95d'  # Форма элемента справочника
    ChartOfAccounts = '238e7e88-3c5f-48b2-8a3b-81ebbecb20ed'  # План счетов
    ChartOfAccountsCommand = '0df30176-6865-4787-9fc8-609eb144174f'  # План счетов команды include
    ChartOfAccountsForm = '5372e285-03db-4f8c-8565-fe56f1aea40e'  # План счетов
    ChartOfCalculationTypes = '30b100d6-b29f-47ac-aec7-cb8ca8a54767'  # План видов расчета
    ChartOfCalculationTypesCommand = "2e90c75b-2f0c-4899-a7d4-5426eaefc96e"
    ChartOfCalculationTypesForm = 'a7f8f92a-7a4b-484b-937e-42d242e64144'
    ChartOfCharacteristicType = '82a1b659-b220-4d94-a9bd-14d757b95a48'  # План видов характеристик
    ChartOfCharacteristicTypeCommand = "95b5e1d4-abfa-4a16-818d-a5b07b7d3f73"
    ChartOfCharacteristicTypeForm = 'eb2b78a8-40a6-4b7e-b1b3-6ca9966cbc94'  # План видов характеристик
    CommandGroup = '1c57eabe-7349-44b3-b1de-ebfeab67b47d'  # Группа команд
    CommonAttribute = '15794563-ccec-41f6-a83c-ec5f7b9a5bc1'  # реквизиты
    CommonCommand = '2f1a5187-fb0e-4b05-9489-dc5dd6412348'  # Общая команда
    CommonForm = "07ee8426-87f1-11d5-b99c-0050bae0a95d"  # Общие формы
    CommonModule = '0fe48980-252d-11d6-a3c7-0050bae0a776'  # Общие модули
    CommonPicture = '7dcd43d9-aca5-4926-b549-1842e6a4e8cf'  # Общие картинки
    CommonTemplate = '0c89c792-16c3-11d5-b96b-0050bae0a95d'  # Общие макеты
    Configuration = '9cd510cd-abfc-11d4-9434-004095e12fc7'  # Идентифицирует род. узел описания конфигурации
    Constant = '0195e80c-b157-11d4-9435-004095e12fc7'  # Константы
    DataProcessor = "bf845118-327b-4682-b5c6-285d2a0eb296"  # Обработки
    DataProcessorCommand = '45556acb-826a-4f73-898a-6025fc9536e1'
    DefinedType = 'c045099e-13b9-4fb6-9d50-fca00202971e'  # Определяемые типы
    Document = "061d872a-5787-460e-95ac-ed74ea3a3e84"  # Документы
    DocumentCommand = 'b544fc6a-2ba3-4885-8fb2-cb289fb6d65e'  # Команда документа
    DocumentForm = "fb880e93-47d7-4127-9357-a20e69c17545"  # Форма документа
    DocumentJournal = '4612bd75-71b7-4a5c-8cc5-2b0b65f9fa0d'  # Журналы документов
    DocumentJournalCommand = "a49a35ce-120a-4c80-8eea-b0618479cd70"
    DocumentJournalForm = 'ec81ad10-ca07-11d5-b9a5-0050bae0a95d'
    DocumentNumerators = '36a8e346-9aaa-4af9-bdbd-83be3c177977'  # Нумератор документа
    Enum = "f6a80749-5ad7-400b-8519-39dc5dff2542"  # Перечисления
    EnumCommand = "6d8d73a7-ba29-401d-9032-3872ec2d6433"
    EnumForm = "33f2e54b-37ce-4a7a-a569-b648d7aa4634"  # Перечисления
    EventSubscription = '4e828da6-0f44-4b5b-b1c0-a2b3cfe7bdcc'  # Подписки на события
    ExchangePlan = '857c4a91-e5f4-4fac-86ec-787626f1c108'  # План обмена
    ExchangePlanCommand = 'd5207c64-11d5-4d46-bba2-55b7b07ff4eb'
    ExchangePlanForm = '87c509ab-3d38-4d67-b379-aca796298578'  # План обмена Form
    ExternalDataProcessor = 'c3831ec8-d8d5-4f93-8a22-f9bfae07327f'  # Внешняя обработка
    ExternalDataSource = '5274d9fc-9c3a-4a71-8f5e-a0db8ab23de5'  # ВнешнийИсточникДанных
    ExternalDataSourceCube = '2bb208ca-e441-4023-9ab9-32a7807a85d0'  # ВнешнийИсточникДанных Куб
    ExternalDataSourceCubeForm = '3448e506-5add-4fce-a604-7305466b2d8e'  # ВнешнийИсточникДанных Куб
    ExternalDataSourceCubeCommand = '4ee40ec7-3469-439f-adb4-aa26ce2d3ec3'  # ВнешнийИсточникДанных Куб
    ExternalDataSourceTable = 'e3403acd-1c95-421b-87e4-4dfa29d38b52'  # ВнешнийИсточникДанных Таблица
    ExternalDataSourceTableForm = '17816ebc-4068-496e-adc4-8879945a832f'  # ВнешнийИсточникДанных Таблица
    ExternalDataSourceTableCommand = '5bb6f09e-5d80-41f6-8070-9faa4d15b69b'  # ВнешнийИсточникДанных Таблица
    FilterCriterion = '3e7bfcc0-067d-11d6-a3c7-0050bae0a776'  # Критерии отбора
    FilterCriterionCommand = "23fa3b84-220a-40e9-8331-e588bed87f7d"
    FilterCriterionForm = '00867c40-06b1-11d6-a3c7-0050bae0a776'  # Критерии отбора Form
    Form = 'd5b0e5ed-256d-401c-9c36-f630cafd8a62'  # Форма внешней обработки
    FormAttribute = 'ec6bb5e5-b7a8-4d75-bec9-658107a699cf'  # Атрибут Внешней обработки
    FunctionalOption = 'af547940-3268-434f-a3e7-e47d6d2638c3'  # Функциональные опции
    FunctionalOptionsParameter = '30d554db-541e-4f62-8970-a1c6dcfeb2bc'  # Параметры функциональных опций
    HTTPService = '0fffc09c-8f4c-47cc-b41c-8d5c5a221d79'  # HTTPСервис
    InformationRegister = '13134201-f60b-11d5-a3c7-0050bae0a776'  # Регистры сведений
    InformationRegisterCommand = 'b44ba719-945c-445c-8aab-1088fa4df16e'
    InformationRegisterForm = '13134204-f60b-11d5-a3c7-0050bae0a776'  # Form регистра сведений
    IntegrationService = 'bf3420b0-f6f9-41a0-b83a-fe9d4ab0b65d'#e3403acd-1c95 -421b-87e4-4dfa29d38b52'  # Сервис интеграции
    Interface = '39bddf6a-0c3c-452b-921c-d99cfa1c2f1b'  # Интерфейсы
    Language = '9cd510ce-abfc-11d4-9434-004095e12fc7'  # Языки
    Report = '631b75a0-29e2-11d6-a3c7-0050bae0a776'  # Отчеты
    ReportCommand = 'e7ff38c0-ec3c-47a0-ae90-20c73ca72246'  # Отчеты
    ReportForm = 'a3b368c0-29e2-11d6-a3c7-0050bae0a776'  # Отчеты форма
    Role = '09736b02-9cac-4e3f-b4f7-d3e9576ab948'  # Роли
    ScheduledJob = '11bdaf85-d5ad-4d91-bb24-aa0eee139052'  # Регламентные задания
    Sequences = 'bc587f20-35d9-11d6-a3c7-0050bae0a776'  # Последовательность, 1 ШТ
    SessionParameter = '24c43748-c938-45d0-8d14-01424a72b11e'  # Параметры сеанса
    SettingsStorage = '46b4cd97-fd13-4eaa-aba2-3bddd7699218'  # Хранилища настроек
    SettingsStorageForm = 'b8533c0c-2342-4db3-91a2-c2b08cbf6b23'
    Style = '3e5404af-6ef8-4c73-ad11-91bd2dfac4c8'  # Стиль
    StyleItem = '58848766-36ea-4076-8800-e91eb49590d7'  # Элемены стиля
    Subsystem = '37f2fa9a-b276-11d4-9435-004095e12fc7'  # Подсистемы
    TabularSectionAttribute = '2bcef0d1-0981-11d6-b9b8-0050bae0a95d'  # Атрибут Табличной части Внешней обработки
    Task = '3e63355c-1378-4953-be9b-1deb5fb6bec5'  # Задачи
    TaskCommand = 'f27c2152-a2c9-4c30-adb1-130f5eb2590f'
    TaskForm = '3f58cbfb-4172-4e54-be49-561a579bb38b'
    Template = '3daea016-69b7-4ed4-9453-127911372fe6'  # Макет внешней обработки
    WebService = '8657032e-7740-4e1d-a3ba-5dd6e8afb78f'  # Веб-сервис
    WSReference = 'd26096fb-7a5d-4df9-af63-47d04771fa9b'  # WS-ссылки
    XDTOPackage = 'cc9df798-7c94-4616-97d2-7aa0b7bc515e'  # ПакетXDTO








import os
import shutil

from ..MetaDataObject.core.Simple import Simple


class WSReference(Simple):
    ext_code = {}

    @classmethod
    def get_decode_header(cls, header):
        return header[0][1][2]

    def decode_object(self, src_dir, file_name, dest_dir, dest_path, version, header_data):
        super().decode_object(src_dir, file_name, dest_dir, dest_path, version, header_data)

        src = os.path.join(src_dir, f'{self.header["uuid"]}.0')
        dest = os.path.join(dest_dir, dest_path, self.header["name"])
        if not os.path.isdir(src):
            return
        shutil.copytree(src, dest)

    def encode_object(self, src_dir, file_name, dest_dir):
        super().encode_object(src_dir, file_name, dest_dir)

        src = os.path.join(src_dir, self.header["name"])
        if not os.path.isdir(src):
            return
        dir_name = f'{self.header["uuid"]}.0'
        dest = os.path.join(dest_dir, dir_name)
        shutil.copytree(src, dest)
        self.file_list.append(dir_name)



# -*- coding: utf-8 -*-
import datetime
import os
import tempfile
from base64 import b64encode
from datetime import datetime
from hashlib import sha1

from tqdm.auto import tqdm

from . import helper
from .container import Container, Container64
from .container_doc import Document
from .json_container_decoder import JsonContainerDecoder


def build(src_dir, filename, nested=False):
    """
    Запаковывает каталог в контейнер включая вложенные каталоги.
    Сахар для ContainerWriter.

    :param src_dir: каталог с данными, запаковываемыми в контейнер
    :type src_dir: string
    :param filename: имя файла контейнера
    :type filename: string
    :param nested:
    :type nested: bool
    :param version: версия под которую собирается продукт
    :type version: string
    """
    begin = datetime.now()
    print(f'{"Запаковываем бинарник":30}:', end="")
    helper.makedirs(os.path.dirname(filename), exist_ok=True)

    containers = os.listdir(src_dir)
    _src_dir = containers[-1]
    containers_count = len(containers)
    if containers_count not in [1, 2]:
        raise NotImplementedError(f'Количество контейнеров {containers_count}')

    with open(filename, 'w+b') as f:
        _src_dir = os.path.join(src_dir, '0')
        container = Container()
        container.build(f, _src_dir, nested, offset=0)
        if containers_count == 2:
            _src_dir = os.path.join(src_dir, '1')
            container = Container64()
            container.build(f, _src_dir, nested, offset=f.seek(0, os.SEEK_END))
    print(f" - {datetime.now() - begin}")


def calc_sha1(src_folder, dest_folder):
    entries = sorted(os.listdir(dest_folder))
    versions = []
    versions_file = ''
    for filename in entries:
        dest_path = os.path.join(dest_folder, filename)
        if filename == 'configinfo':
            versions_file = 'configinfo'
            continue

        with open(dest_path, 'rb') as file:
            data = file.read()
        versions.append(f'"{filename}"')
        versions.append(b64encode(sha1(data).digest()).decode())
    if not versions_file:
        return
    versions.insert(0, str(int(len(versions) / 2)))
    _versions = JsonContainerDecoder().encode_root_object([versions])

    src_path = os.path.join(src_folder, versions_file)
    dest_path = os.path.join(dest_folder, versions_file)

    with open(src_path, 'r+', encoding='utf-8') as file:
        data = file.read()
        data = data.replace('{____versions____}', _versions)
        file.seek(0)
        file.write(data)

    compress_and_build_simple_file(src_path, dest_path)
    pass


def compress_and_build_simple_file(src_path, dest_path):
    with open(dest_path, 'w+b') as dest_fd:
        with open(src_path, 'rb') as src_fd:
            Document.compress(src_fd, dest_fd)


def compress_and_build(src_dir, dest_dir, *, pool=None, nested=False):
    containers = os.listdir(src_dir)
    helper.clear_dir(dest_dir)
    for dir_name in containers:
        _src_dir = os.path.join(src_dir, dir_name)
        _dest_dir = os.path.join(dest_dir, dir_name)
        helper.clear_dir(_dest_dir)

        entries = sorted(os.listdir(_src_dir))

        with tqdm(desc=f'{"Архивируем контейнеры":30}', total=len(entries)) as pbar:
            for filename in entries:
                src_path = os.path.join(_src_dir, filename)
                dest_path = os.path.join(_dest_dir, filename)
                # add_entry(container, src_dir, filename)
                if os.path.isdir(src_path):
                    with open(dest_path, 'w+b') as dest_fd:
                        with tempfile.TemporaryFile() as tmp:
                            container = Container()
                            container.build(tmp, src_path, nested=True)
                            Document.compress(tmp, dest_fd)
                else:
                    compress_and_build_simple_file(src_path, dest_path)
                pbar.update()
            calc_sha1(_src_dir, _dest_dir)



import os
import shutil
import sys
import unittest

sys.path.append("../../src/")
from v8unpack import helper
from v8unpack.file_organizer_ce import FileOrganizerCE


class TestFileOrganizerCE(unittest.TestCase):
    def setUp(self) -> None:
        self.current_dir = os.path.dirname(__file__)
        self.data_dir = os.path.join(self.current_dir, 'data')
        self.temp_dir = os.path.join(self.data_dir, 'temp')

    def test_near(self):
        helper.clear_dir(self.temp_dir)
        self.copy_data_to_temp('test.obj.1c', ['3000070060', '3000075100', '3000088001'])

        _path, _file_name = helper.get_near_descent_file_name(self.temp_dir, 'test.obj.1c', 3000075100)
        self.assertEqual(self.temp_dir, _path)
        self.assertEqual('test.obj.3000075100.1c', _file_name)
        _path, _file_name = helper.get_near_descent_file_name(self.temp_dir, 'test.obj.1c', 3000088003)
        self.assertEqual(self.temp_dir, _path)
        self.assertEqual('test.obj.3000088001.1c', _file_name)
        _path, _file_name = helper.get_near_descent_file_name(self.temp_dir, 'test.obj.1c', 1)
        self.assertEqual('', _path)
        self.assertEqual('', _file_name)
        _path, _file_name = helper.get_near_descent_file_name(self.temp_dir, 'test.2.obj.1c', 1)
        self.assertEqual('', _path)
        self.assertEqual('', _file_name)

    def test_1(self):
        helper.clear_dir(self.temp_dir)
        self.copy_data_to_temp('test.1c', ['3000070060', '3000075100', '3000088001'], '')

        FileOrganizerCE.unpack_file(self.data_dir, 'test.1c', self.temp_dir, '', 'test.1c', {}, 3000088001)

        self.assertEqual(
            os.path.getsize(os.path.join(self.data_dir, 'test.1c')),
            os.path.getsize(os.path.join(self.temp_dir, 'test.3000088001.1c'))
        )

        FileOrganizerCE.unpack_file(self.data_dir, 'test.1c', self.temp_dir, '', 'test.1c', {}, 3000075101)
        self.assertEqual(
            os.path.getsize(os.path.join(self.data_dir, 'test.1c')),
            os.path.getsize(os.path.join(self.temp_dir, 'test.3000075101.1c'))
        )

        self.copy_data_to_temp('test.obj.1c', ['3000070060'])
        FileOrganizerCE.unpack_file(self.data_dir, 'test.obj.1c', self.temp_dir, '', 'test.obj.1c', {}, 3000070061)
        self.assertRaises(
            FileNotFoundError,
            lambda: os.path.getsize(os.path.join(self.temp_dir, 'test.obj.3000070061.1c'))
        )
        self.assertEqual(
            os.path.getsize(os.path.join(self.data_dir, 'test.obj.1c')),
            os.path.getsize(os.path.join(self.temp_dir, 'test.obj.3000070060.1c'))
        )

    def copy_data_to_temp(self, src_file_name, descents: list, value=None):
        for descent in descents:
            dest_file_name = helper.get_descent_file_name(src_file_name, descent)
            dest_full_path = os.path.join(self.temp_dir, dest_file_name)
            shutil.copy(
                os.path.join(self.data_dir, src_file_name),
                dest_full_path
            )
            if value is not None:
                with open(dest_full_path, 'w', encoding='utf-8') as file:
                    file.write('')



from ..MetaDataObject.core.Container import Container
from .. import helper


class ChartOfAccounts(Container):
    ext_code = {
        'obj': 14,
        'mgr': 15
    }
    help_file_number = 5
    predefined_file_number = 9

    @classmethod
    def get_decode_header(cls, header):
        return header[0][1][15][1]

    # @classmethod
    # def get_decode_includes(cls, header_data):
    #     return super().get_decode_includes(header_data)


import os
import sys
import unittest

sys.path.append("../../src/")
from v8unpack import helper
from v8unpack.MetaDataObject.versions.Form803 import Form803
from v8unpack.MetaDataObject.versions.Form802 import Form802
from v8unpack.unittest_helper import compare_file, NotEqualLine


class TestFormElem803(unittest.TestCase):
    def setUp(self) -> None:
        self.current_dir = os.path.dirname(__file__)
        self.data_dir = os.path.join(self.current_dir, 'data', 'form_elem_803')
        self.temp_dir = os.path.join(self.data_dir, 'temp')
        self.temp_dir2 = os.path.join(self.data_dir, 'temp2')

    def test_decode_all_form_elem(self):
        test_files = os.listdir(self.data_dir)
        result = ''
        for file_name in test_files:
            if file_name == 'temp':
                continue
            item = file_name.split('.')
            result += self.decode_form_elem(item[0])
        self.assertEqual('', result)

    def test_decode_form_elem(self):
        uuid = '66f22eef-9167-4661-acc1-769de41ab428'
        result = self.decode_form_elem(uuid)
        self.assertEqual('', result)

    def decode_form_elem(self, uuid):
        file_name = f'{uuid}.0'
        json_file_name = f'{file_name}.json'
        helper.clear_dir(self.temp_dir)
        if not os.path.isfile(os.path.join(self.data_dir, json_file_name)):
            raw_data = helper.brace_file_read(self.data_dir, file_name)
            helper.json_write(raw_data, self.data_dir, json_file_name)
            os.remove(os.path.join(self.data_dir, file_name))
        else:
            raw_data = helper.json_read(self.data_dir, json_file_name)
        form = Form802() if raw_data[0][0] == '2' else Form803()
        form.new_dest_dir = self.temp_dir
        form.form = [raw_data]
        form.decode_includes(None, self.temp_dir, '', None)

        form.write_decode_object(self.temp_dir, '', uuid)
        form.encode_nested_includes(self.temp_dir, uuid, self.temp_dir, '')
        helper.json_write(form.form[0], self.temp_dir, json_file_name)
        problems = ''
        try:
            result = compare_file(
                os.path.join(self.data_dir, json_file_name),
                os.path.join(self.temp_dir, json_file_name),
                problems
            )
        except NotEqualLine as err:
            result = str(err)
        if result:
            print(result)
        return result



from ..MetaDataObject.core.Container import Container


class DataProcessor(Container):
    version = '803'
    ext_code = {
        'mgr': '2',  # модуль менеджера
        'obj': '0',  # модуль объекта
    }
    help_file_number = 1

    def __init__(self, *, obj_name=None, options=None):
        super().__init__(obj_name=obj_name, options=options)

    @classmethod
    def get_decode_header(cls, header_data):
        return header_data[0][1][3][1]



import os

from ..MetaDataObject.core.Simple import SimpleNameFolder
from ..MetaDataObject.versions.Form801 import Form801
from ..MetaDataObject.versions.Form802 import Form802
from ..MetaDataObject.versions.Form803 import Form803
from ..MetaDataObject.versions.OldForm803 import OldForm801, OldForm802, OldForm803
from ..MetaDataObject.versions.Form802 import Form802
from ..ext_exception import ExtException


class Form(SimpleNameFolder):
    versions = {
        '801': Form801,
        '802': Form802,
        '803': Form803,
        '5-5': OldForm801,
        '7-7': OldForm801,
        '9-9': OldForm803,
        '12-12': OldForm803,
        '13-13': OldForm803,
        '0-5': Form801,
        '0-7': Form801,
        '0-9': Form803,
        '0-12': Form803,
        '0-13': Form803,
        '1-9': Form803,
        '1-12': Form803,
        '1-13': Form803
    }
    form_version_index = {
        "0": 3,
        "1": 4,
        "5": 2,
        "7": 2,
        "9": 2,
        "12": 2,
        "13": 2,
    }
    @classmethod
    def decode_get_handler(cls, src_dir, file_name, options):
        obj_version, form_version = cls.get_form_version(src_dir, file_name)
        handler = cls.get_version(f'{obj_version}-{form_version}')(options=options)
        handler.obj_version = obj_version
        return handler

    @classmethod
    def get_form_version(cls, path, file_name):
        try:
            _path = os.path.join(path, file_name)
            with open(_path, 'r', encoding='utf-8-sig') as file:
                header = file.read(20)
                root = header.split('{')

                obj_type_version, x = root[2].split(',')
                form_version, x = root[cls.form_version_index[obj_type_version]].split(',')
                return obj_type_version, form_version
        except Exception as err:
            raise ExtException(message="Неудалось опеределить версию формы", detail=f'{cls.__name__} {file_name}')




import os
import sys
import unittest

sys.path.append("../../src/")
from v8unpack.container_reader import extract as container_extract, decompress_and_extract
from v8unpack.container_writer import build as container_build, compress_and_build


class TestFileOrganizerCE(unittest.TestCase):
    def setUp(self) -> None:
        self.current_dir = os.path.dirname(__file__)
        self.data_dir = os.path.join(self.current_dir, 'data')
        self.temp_dir = os.path.join(self.data_dir, 'temp')

    def test_extract_old(self):
        src_filename = os.path.join(self.data_dir, 'apam_old.cf')
        dest_dir0 = os.path.join(self.temp_dir, 'apam-0')
        dest_dir1 = os.path.join(self.temp_dir, 'apam-1')
        container_extract(src_filename, dest_dir0, False, False)
        decompress_and_extract(dest_dir0, dest_dir1)

    def test_extract_16(self):
        src_filename = os.path.join(self.data_dir, 'apam.cf')
        dst_filename = os.path.join(self.temp_dir, 'apam.cf')
        # src_filename = os.path.join(self.data_dir, 'apam_old.cf')
        # src_filename = os.path.join(self.data_dir, '1Cv8_8316.cf')
        dest_dir0 = os.path.join(self.temp_dir, 'apam16-0')
        dest_dir1 = os.path.join(self.temp_dir, 'apam16-1')
        dest_dir2 = os.path.join(self.temp_dir, 'apam16-2')
        container_extract(src_filename, dest_dir0, False, False)
        decompress_and_extract(dest_dir0, dest_dir1)
        compress_and_build(dest_dir1, dest_dir2)
        container_build(dest_dir2, dst_filename, True, version='80316')

#[(559, 686), (2451, 2578), (3121, 3184), (3727, 3796), (4339, 4410)]